#
# Copyright (c) 2015 Matthew Bentley
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import hashlib

import duplicity.backend
from duplicity.errors import BackendException, FatalBackendException
from duplicity import log
from duplicity import progress


class B2ProgressListener:
    def set_total_bytes(self, total_byte_count):
        self.total_byte_count = total_byte_count

    def bytes_completed(self, byte_count):
        progress.report_transfer(byte_count, self.total_byte_count)

    def close(self):
        pass


class B2Backend(duplicity.backend.Backend):
    """
    Backend for BackBlaze's B2 storage service
    """

    def __init__(self, parsed_url):
        """
        Authorize to B2 api and set up needed variables
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        # Import B2 API
        try:
            global b2
            import b2
            import b2.api
            import b2.account_info
            import b2.download_dest
            import b2.file_version
        except ImportError:
            raise BackendException('B2 backend requires B2 Python APIs (pip install b2)')

        self.service = b2.api.B2Api(b2.account_info.InMemoryAccountInfo())
        self.parsed_url.hostname = 'B2'

        account_id = parsed_url.username
        account_key = self.get_password()

        self.url_parts = [
            x for x in parsed_url.path.replace("@", "/").split('/') if x != ''
        ]
        if self.url_parts:
            self.username = self.url_parts.pop(0)
            bucket_name = self.url_parts.pop(0)
        else:
            raise BackendException("B2 requires a bucket name")
        self.path = "".join([url_part + "/" for url_part in self.url_parts])
        self.service.authorize_account('production', account_id, account_key)

        log.Log("B2 Backend (path= %s, bucket= %s, minimum_part_size= %s)" %
                (self.path, bucket_name, self.service.account_info.get_minimum_part_size()), log.INFO)
        try:
            self.bucket = self.service.get_bucket_by_name(bucket_name)
            log.Log("Bucket found", log.INFO)
        except b2.exception.NonExistentBucket:
            try:
                log.Log("Bucket not found, creating one", log.INFO)
                self.bucket = self.service.create_bucket(bucket_name, 'allPrivate')
            except:
                raise FatalBackendException("Bucket cannot be created")

    def _get(self, remote_filename, local_path):
        """
        Download remote_filename to local_path
        """
        log.Log("Get: %s -> %s" % (self.path + remote_filename, local_path.name), log.INFO)
        self.bucket.download_file_by_name(self.path + remote_filename,
                                          b2.download_dest.DownloadDestLocalFile(local_path.name))

    def _put(self, source_path, remote_filename):
        """
        Copy source_path to remote_filename
        """
        log.Log("Put: %s -> %s" % (source_path.name, self.path + remote_filename), log.INFO)
        self.bucket.upload_local_file(source_path.name, self.path + remote_filename,
                                      content_type='application/pgp-encrypted',
                                      progress_listener=B2ProgressListener())

    def _list(self):
        """
        List files on remote server
        """
        return [file_version_info.file_name[len(self.path):]
                for (file_version_info, folder_name) in self.bucket.ls(self.path)]

    def _delete(self, filename):
        """
        Delete filename from remote server
        """
        log.Log("Delete: %s" % self.path + filename, log.INFO)
        file_version_info = self.file_info(self.path + filename)
        self.bucket.delete_file_version(file_version_info.id_, file_version_info.file_name)

    def _query(self, filename):
        """
        Get size info of filename
        """
        log.Log("Query: %s" % self.path + filename, log.INFO)
        file_version_info = self.file_info(self.path + filename)
        return {'size': file_version_info.size
                if file_version_info is not None and file_version_info.size is not None else -1}

    def file_info(self, filename):
        response = self.bucket.list_file_names(filename, 1)
        for entry in response['files']:
            file_version_info = b2.file_version.FileVersionInfoFactory.from_api_response(entry)
            if file_version_info.file_name == filename:
                return file_version_info
        raise BackendException('File not found')


duplicity.backend.register_backend("b2", B2Backend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2014 Michael Terry <michael.terry@canonical.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import os
import sys
import subprocess

if sys.version_info < (2, 7):
    import unittest2 as unittest  # @UnresolvedImport @UnusedImport
else:
    import unittest  # @Reimport

from . import _top_dir, DuplicityTestCase  # @IgnorePep8


class CodeTest(DuplicityTestCase):

    def run_checker(self, cmd, returncodes=[0]):
        process = subprocess.Popen(cmd,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        output = process.communicate()[0]
        self.assertTrue(process.returncode in returncodes, output)
        self.assertEqual("", output, output)

    @unittest.skipUnless(os.getenv('RUN_CODE_TESTS', None) == '1',
                         'Must set environment var RUN_CODE_TESTS=1')
    def test_2to3(self):
        # As we modernize the source code, we can remove more and more nofixes
        self.run_checker([
            "2to3",
            "--nofix=next",
            "--nofix=types",
            "--nofix=unicode",
            # The following fixes we don't want to remove, since they are false
            # positives, things we don't care about, or real incompatibilities
            # but which 2to3 can fix for us better automatically.
            "--nofix=callable",
            "--nofix=dict",
            "--nofix=future",
            "--nofix=imports",
            "--nofix=print",
            "--nofix=raw_input",
            "--nofix=urllib",
            "--nofix=xrange",
            _top_dir])

    @unittest.skipUnless(os.getenv('RUN_CODE_TESTS', None) == '1',
                         'Must set environment var RUN_CODE_TESTS=1')
    def test_pylint(self):
        """Pylint test (requires pylint to be installed to pass)"""
        self.run_checker([
            "pylint",
            "-E",
            "--msg-template={msg_id}: {line}: {msg}",
            "--disable=E0203",  # Access to member before its definition line
            "--disable=E0602",  # Undefined variable
            "--disable=E0611",  # No name in module
            "--disable=E1101",  # Has no member
            "--disable=E1103",  # Maybe has no member
            "--disable=E0712",  # Catching an exception which doesn't inherit from BaseException
            "--ignore=_librsync.so",
            "--msg-template='{path}:{line}: [{msg_id}({symbol}), {obj}] {msg}'",
            os.path.join(_top_dir, 'duplicity'),
            os.path.join(_top_dir, 'bin/duplicity'),
            os.path.join(_top_dir, 'bin/rdiffdir')],
            # Allow usage errors, older versions don't have
            # --msg-template
            [0, 32])

    @unittest.skipUnless(os.getenv('RUN_CODE_TESTS', None) == '1',
                         'Must set environment var RUN_CODE_TESTS=1')
    def test_pep8(self):
        ignores = [
            "E402",  # module level import not at top of file
            "E731",  # do not assign a lambda expression, use a def
        ]
        self.run_checker(["pep8",
                          "--ignore=" + ','.join(ignores),
                          "--max-line-length=120",
                          os.path.join(_top_dir, 'duplicity'),
                          os.path.join(_top_dir, 'bin/duplicity'),
                          os.path.join(_top_dir, 'bin/rdiffdir')])


if __name__ == "__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import os
import sys
import random
import unittest

from duplicity import collections
from duplicity import backend
from duplicity import globals
from duplicity import path
from duplicity import gpg
from duplicity import dup_time
from . import UnitTestCase

filename_list1 = ["duplicity-full.2002-08-17T16:17:01-07:00.manifest.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.vol1.difftar.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.vol2.difftar.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.vol3.difftar.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.vol4.difftar.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.vol5.difftar.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.vol6.difftar.gpg",
                  "duplicity-inc.2002-08-17T16:17:01-07:00.to.2002-08-18T00:04:30-07:00.manifest.gpg",
                  "duplicity-inc.2002-08-17T16:17:01-07:00.to.2002-08-18T00:04:30-07:00.vol1.difftar.gpg",
                  "Extra stuff to be ignored"]

remote_sigchain_filename_list = ["duplicity-full-signatures.2002-08-17T16:17:01-07:00.sigtar.gpg",
                                 "duplicity-new-signatures.2002-08-17T16:17:01-07:00.to.2002-08-18T00:04:30-07:00.sigtar.gpg",
                                 "duplicity-new-signatures.2002-08-18T00:04:30-07:00.to.2002-08-20T00:00:00-07:00.sigtar.gpg"]

local_sigchain_filename_list = ["duplicity-full-signatures.2002-08-17T16:17:01-07:00.sigtar.gz",
                                "duplicity-new-signatures.2002-08-17T16:17:01-07:00.to.2002-08-18T00:04:30-07:00.sigtar.gz",
                                "duplicity-new-signatures.2002-08-18T00:04:30-07:00.to.2002-08-20T00:00:00-07:00.sigtar.gz"]

# A filename list with some incomplete volumes, an older full volume,
# and a complete chain.
filename_list2 = ["duplicity-full.2001-01-01T16:17:01-07:00.manifest.gpg",
                  "duplicity-full.2001-01-01T16:17:01-07:00.vol1.difftar.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.manifest.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.vol1.difftar.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.vol2.difftar.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.vol3.difftar.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.vol4.difftar.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.vol5.difftar.gpg",
                  "duplicity-full.2002-08-17T16:17:01-07:00.vol6.difftar.gpg",
                  "duplicity-inc.2002-08-17T16:17:01-07:00.to.2002-08-18T00:04:30-07:00.manifest.gpg",
                  "duplicity-inc.2002-08-17T16:17:01-07:00.to.2002-08-18T00:04:30-07:00.vol1.difftar.gpg",
                  "The following are extraneous duplicity files",
                  "duplicity-new-signatures.2001-08-17T02:05:13-05:00.to.2002-08-17T05:05:14-05:00.sigtar.gpg",
                  "duplicity-full.2002-08-15T01:01:01-07:00.vol1.difftar.gpg",
                  "duplicity-inc.2000-08-17T16:17:01-07:00.to.2000-08-18T00:04:30-07:00.manifest.gpg",
                  "duplicity-inc.2000-08-17T16:17:01-07:00.to.2000-08-18T00:04:30-07:00.vol1.difftar.gpg",
                  "Extra stuff to be ignored"]


class CollectionTest(UnitTestCase):
    """Test collections"""
    def setUp(self):
        super(CollectionTest, self).setUp()

        self.unpack_testfiles()

        col_test_dir = path.Path("testfiles/collectionstest")
        archive_dir = col_test_dir.append("archive_dir")
        self.set_global('archive_dir', archive_dir)
        self.archive_dir_backend = backend.get_backend("file://testfiles/collectionstest"
                                                       "/archive_dir")

        self.real_backend = backend.get_backend("file://%s/%s" %
                                                (col_test_dir.name, "remote_dir"))
        self.output_dir = path.Path("testfiles/output")  # used as a temp directory
        self.output_dir_backend = backend.get_backend("file://testfiles/output")

    def set_gpg_profile(self):
        """Set gpg profile to standard "foobar" sym"""
        self.set_global('gpg_profile', gpg.GPGProfile(passphrase="foobar"))

    def test_backup_chains(self):
        """Test basic backup chain construction"""
        random.shuffle(filename_list1)
        cs = collections.CollectionsStatus(None, globals.archive_dir, "full")
        chains, orphaned, incomplete = cs.get_backup_chains(filename_list1)  # @UnusedVariable
        if len(chains) != 1 or len(orphaned) != 0:
            print chains
            print orphaned
            assert 0

        chain = chains[0]
        assert chain.end_time == 1029654270
        assert chain.fullset.time == 1029626221

    def test_collections_status(self):
        """Test CollectionStatus object's set_values()"""
        def check_cs(cs):
            """Check values of collections status"""
            assert cs.values_set

            assert cs.matched_chain_pair
            assert cs.matched_chain_pair[0].end_time == 1029826800
            assert len(cs.all_backup_chains) == 1, cs.all_backup_chains

        cs = collections.CollectionsStatus(self.real_backend, globals.archive_dir, "full").set_values()
        check_cs(cs)
        assert cs.matched_chain_pair[0].islocal()

    def test_sig_chain(self):
        """Test a single signature chain"""
        chain = collections.SignatureChain(1, globals.archive_dir)
        for filename in local_sigchain_filename_list:
            assert chain.add_filename(filename)
        assert not chain.add_filename("duplicity-new-signatures.2002-08-18T00:04:30-07:00.to.2002-08-20T00:00:00-07:00.sigtar.gpg")

    def test_sig_chains(self):
        """Test making signature chains from filename list"""
        cs = collections.CollectionsStatus(None, globals.archive_dir, "full")
        chains, orphaned_paths = cs.get_signature_chains(local=1)
        self.sig_chains_helper(chains, orphaned_paths)

    def test_sig_chains2(self):
        """Test making signature chains from filename list on backend"""
        cs = collections.CollectionsStatus(self.archive_dir_backend, globals.archive_dir, "full")
        chains, orphaned_paths = cs.get_signature_chains(local=None)
        self.sig_chains_helper(chains, orphaned_paths)

    def sig_chains_helper(self, chains, orphaned_paths):
        """Test chains and orphaned_paths values for two above tests"""
        if orphaned_paths:
            for op in orphaned_paths:
                print op
            assert 0
        assert len(chains) == 1, chains
        assert chains[0].end_time == 1029826800

    def sigchain_fileobj_get(self, local):
        """Return chain, local if local is true with filenames added"""
        if local:
            chain = collections.SignatureChain(1, globals.archive_dir)
            for filename in local_sigchain_filename_list:
                assert chain.add_filename(filename)
        else:
            chain = collections.SignatureChain(None, self.real_backend)
            for filename in remote_sigchain_filename_list:
                assert chain.add_filename(filename)
        return chain

    def sigchain_fileobj_check_list(self, chain):
        """Make sure the list of file objects in chain has right contents

        The contents of the testfiles/collectiontest/remote_dir have
        to be coordinated with this test.

        """
        fileobjlist = chain.get_fileobjs()
        assert len(fileobjlist) == 3

        def test_fileobj(i, s):
            buf = fileobjlist[i].read()
            fileobjlist[i].close()
            assert buf == s, (buf, s)

        test_fileobj(0, "Hello, world!")
        test_fileobj(1, "hello 1")
        test_fileobj(2, "Hello 2")

    def test_sigchain_fileobj(self):
        """Test getting signature chain fileobjs from archive_dir"""
        self.set_gpg_profile()
        self.sigchain_fileobj_check_list(self.sigchain_fileobj_get(1))
        self.sigchain_fileobj_check_list(self.sigchain_fileobj_get(None))

    def get_filelist2_cs(self):
        """Return set CollectionsStatus object from filelist 2"""
        # Set up testfiles/output with files from filename_list2
        for filename in filename_list2:
            p = self.output_dir.append(filename)
            p.touch()

        cs = collections.CollectionsStatus(self.output_dir_backend, globals.archive_dir, "full")
        cs.set_values()
        return cs

    def test_get_extraneous(self):
        """Test the listing of extraneous files"""
        cs = self.get_filelist2_cs()
        assert len(cs.orphaned_backup_sets) == 1, cs.orphaned_backup_sets
        assert len(cs.local_orphaned_sig_names) == 0, cs.local_orphaned_sig_names
        assert len(cs.remote_orphaned_sig_names) == 1, cs.remote_orphaned_sig_names
        assert len(cs.incomplete_backup_sets) == 1, cs.incomplete_backup_sets

        right_list = ["duplicity-new-signatures.2001-08-17T02:05:13-05:00.to.2002-08-17T05:05:14-05:00.sigtar.gpg",
                      "duplicity-full.2002-08-15T01:01:01-07:00.vol1.difftar.gpg",
                      "duplicity-inc.2000-08-17T16:17:01-07:00.to.2000-08-18T00:04:30-07:00.manifest.gpg",
                      "duplicity-inc.2000-08-17T16:17:01-07:00.to.2000-08-18T00:04:30-07:00.vol1.difftar.gpg"]
        local_received_list, remote_received_list = cs.get_extraneous(False)  # @UnusedVariable
        errors = []
        for filename in remote_received_list:
            if filename not in right_list:
                errors.append("### Got bad extraneous filename " + filename)
            else:
                right_list.remove(filename)
        for filename in right_list:
            errors.append("### Didn't receive extraneous filename " + filename)
        assert not errors, "\n" + "\n".join(errors)

    def test_get_olderthan(self):
        """Test getting list of files older than a certain time"""
        cs = self.get_filelist2_cs()
        oldsets = cs.get_older_than(
            dup_time.genstrtotime("2002-05-01T16:17:01-07:00"))
        oldset_times = [s.get_time() for s in oldsets]
        right_times = [dup_time.genstrtotime('2001-01-01T16:17:01-07:00')]
        assert oldset_times == right_times, \
            [oldset_times, right_times]

        oldsets_required = cs.get_older_than_required(
            dup_time.genstrtotime("2002-08-17T20:00:00-07:00"))
        oldset_times = [s.get_time() for s in oldsets_required]
        right_times_required = [dup_time.genstrtotime('2002-08-17T16:17:01-07:00')]
        assert oldset_times == right_times_required, \
            [oldset_times, right_times_required]


if __name__ == "__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
# Copyright 2014 Aaron Whitehouse <aaron@whitehouse.kiwi.nz>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import sys
from duplicity.globmatch import *
from duplicity.path import *
from . import UnitTestCase
from mock import patch
import unittest


class MatchingTest(UnitTestCase):
    """Test matching of file names against various selection functions"""

    def test_glob_re(self):
        """test_glob_re - test translation of shell pattern to regular exp"""
        assert glob_to_regex("hello") == "hello"
        assert glob_to_regex(".e?ll**o") == "\\.e[^/]ll.*o"
        r = glob_to_regex("[abc]el[^de][!fg]h")
        assert r == "[abc]el[^de][^fg]h", r
        r = glob_to_regex("/usr/*/bin/")
        assert r == "\\/usr\\/[^/]*\\/bin\\/", r
        assert glob_to_regex("[a.b/c]") == "[a.b/c]"
        r = glob_to_regex("[a*b-c]e[!]]")
        assert r == "[a*b-c]e[^]]", r


class TestDoubleAsteriskOnIncludesExcludes(UnitTestCase):
    """Test ** on includes and exclude patterns"""

    def test_double_asterisk_include(self):
        """Test a few globbing patterns, including **"""
        self.assertEqual(
            path_matches_glob_fn("**", 1)(Path("foo.txt")), 1)
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = True
            self.assertEqual(
                path_matches_glob_fn("**", 1)(Path("folder")), 1)

    def test_double_asterisk_extension_include(self):
        """Test **.py"""
        self.assertEqual(
            path_matches_glob_fn("**.py", 1)(Path("what/ever.py")), 1)
        self.assertEqual(
            path_matches_glob_fn("**.py", 1)(Path("what/ever.py/foo")), 1)
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = True
            self.assertEqual(
                path_matches_glob_fn("**.py", 1)(Path("foo")), 2)
            self.assertEqual(
                path_matches_glob_fn("**.py", 1)(Path("usr/local/bin")), 2)
            self.assertEqual(
                path_matches_glob_fn("**.py", 1)(Path("/usr/local/bin")), 2)


class TestTrailingSlash(UnitTestCase):
    """Test glob matching where the glob has a trailing slash"""

    def test_trailing_slash_matches_only_dirs(self):
        """Test matching where glob includes a trailing slash"""
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = True
            self.assertEqual(
                path_matches_glob_fn("fold*/", 1)(Path("folder")), 1)

            # Test the file named "folder" is not included if it is not a dir
            mock_isdir.return_value = False
            self.assertEqual(
                path_matches_glob_fn("fold*/", 1)(Path("folder")), None)

            # Test the file named "folder" is not included if it is not a dir
            mock_isdir.return_value = False
            self.assertEqual(
                path_matches_glob_fn("folder/", 1)(Path("folder")), None)

            mock_isdir.return_value = False
            self.assertEqual(
                path_matches_glob_fn("fo*/", 1)(Path("foo.txt")), None)

    def test_included_files_are_matched_no_slash(self):
        """Test that files within an included folder are matched"""
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = False
            self.assertEqual(
                path_matches_glob_fn("fold*", 1)(Path("folder/file.txt")), 1)

        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = False
            self.assertEqual(
                path_matches_glob_fn("fold*", 1)(Path("folder/2/file.txt")), 1)

    def test_included_files_are_matched_no_slash_2(self):
        """Test that files within an included folder are matched"""
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = False
            self.assertEqual(
                path_matches_glob_fn("folder", 1)(Path("folder/file.txt")), 1)

        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = False
            self.assertEqual(
                path_matches_glob_fn("folder/2", 1)(Path("folder/2/file.txt")), 1)

    def test_included_files_are_matched_slash(self):
        """Test that files within an included folder are matched with /"""
        # Bug #1624725
        # https://bugs.launchpad.net/duplicity/+bug/1624725
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = False
            self.assertEqual(
                path_matches_glob_fn("folder/", 1)(Path("folder/file.txt")), 1)

    def test_included_files_are_matched_slash_2(self):
        """Test that files within an included folder are matched with /"""
        # Bug #1624725
        # https://bugs.launchpad.net/duplicity/+bug/1624725
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = False
            self.assertEqual(
                path_matches_glob_fn("testfiles/select2/1/1sub1/1sub1sub1/", 1)
                (Path("testfiles/select2/1/1sub1/1sub1sub1/1sub1sub1_file.txt")
                 ), 1)

    def test_included_files_are_matched_slash_2_parents(self):
        """Test that duplicity will scan parent of glob/"""
        # Bug #1624725
        # https://bugs.launchpad.net/duplicity/+bug/1624725
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = True
            self.assertEqual(
                path_matches_glob_fn("testfiles/select2/1/1sub1/1sub1sub1/", 1)
                (Path("testfiles/select2/1/1sub1/1sub1sub1")
                 ), 1)
            self.assertEqual(
                path_matches_glob_fn("testfiles/select2/1/1sub1/1sub1sub1/", 1)
                (Path("testfiles/select2/1/1sub1")
                 ), 2)

    def test_included_files_are_matched_slash_wildcard(self):
        """Test that files within an included folder are matched with /"""
        # Bug #1624725
        # https://bugs.launchpad.net/duplicity/+bug/1624725
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = False
            self.assertEqual(
                path_matches_glob_fn("fold*/", 1)(Path("folder/file.txt")), 1)
    #
    # def test_slash_matches_everything(self):
    #     """Test / matches everything"""
    #     # ToDo: Not relevant at this stage, as "/" would not go through
    #     # globmatch because it has no special characters, but it should be
    #     # made to work
    #     with patch('duplicity.path.Path.isdir') as mock_isdir:
    #         mock_isdir.return_value = True
    #         self.assertEqual(
    #             path_matches_glob_fn("/",
    #                                  1)(Path("/tmp/testfiles/select/1/2")), 1)
    #         self.assertEqual(
    #             path_matches_glob_fn("/",
    #                                  1)(Path("/test/random/path")), 1)
    #         self.assertEqual(
    #             path_matches_glob_fn("/",
    #                                  1)(Path("/")), 1)
    #         self.assertEqual(
    #             path_matches_glob_fn("/",
    #                                  1)(Path("/var/log")), 1)
    #         self.assertEqual(
    #             path_matches_glob_fn("/",
    #                                  1)(Path("/var/log/log.txt")), 1)

    def test_slash_star_scans_folder(self):
        """Test that folder/* scans folder/"""
        # This behaviour is a bit ambiguous - either include or scan could be
        # argued as most appropriate here, but only an empty folder is at stake
        # so long as test_slash_star_includes_folder_contents passes.
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = True
            self.assertEqual(
                path_matches_glob_fn("folder/*", 1)(Path("folder")), 2)

    def test_slash_star_includes_folder_contents(self):
        """Test that folder/* includes folder contents"""
        self.assertEqual(path_matches_glob_fn("folder/*", 1)
                         (Path("folder/file.txt")), 1)
        self.assertEqual(path_matches_glob_fn("folder/*", 1)
                         (Path("folder/other_file.log")), 1)

    def test_slash_star_star_includes_folder(self):
        """Test that folder/** includes folder/"""
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = True

    def test_simple_trailing_slash_match(self):
        """Test that a normal folder string ending in / matches that path"""
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = True
            test_path = "testfiles/select/1/2/1"
            self.assertEqual(
                path_matches_glob_fn(test_path, 1)(Path(test_path)), 1)

    def test_double_asterisk_string_slash(self):
        """Test string starting with ** and ending in /"""
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = True
            self.assertEqual(
                path_matches_glob_fn("**/1/2/",
                                     1)(Path("testfiles/select/1/2")), 1)

    def test_string_double_asterisk_string_slash(self):
        """Test string ** string /"""
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = True
            self.assertEqual(
                path_matches_glob_fn("testfiles**/2/",
                                     1)(Path("testfiles/select/1/2")), 1)


class TestDoubleAsterisk(UnitTestCase):
    """Test glob matching where the glob finishes with a **"""

    def test_simple_trailing_slash_match(self):
        """Test that a folder string ending in /** matches that path"""
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = True
            self.assertEqual(
                path_matches_glob_fn("/test/folder/**", 1)(
                    Path("/test/foo")), None)

    def test_simple_trailing_slash_match_2(self):
        """Test folder string ending in */**"""
        with patch('duplicity.path.Path.isdir') as mock_isdir:
            mock_isdir.return_value = True
            self.assertEqual(
                path_matches_glob_fn("fold*/**", 1)(
                    Path("folder")), 2)


# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
# Copyright 2013 Edgar Soldin
#                 - ssl cert verification, some robustness enhancements
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import base64
import httplib
import os
import re
import urllib
import urllib2
import urlparse
import xml.dom.minidom

import duplicity.backend
from duplicity import globals
from duplicity import log
from duplicity import util
from duplicity.errors import BackendException, FatalBackendException


class CustomMethodRequest(urllib2.Request):
    """
    This request subclass allows explicit specification of
    the HTTP request method. Basic urllib2.Request class
    chooses GET or POST depending on self.has_data()
    """
    def __init__(self, method, *args, **kwargs):
        self.method = method
        urllib2.Request.__init__(self, *args, **kwargs)

    def get_method(self):
        return self.method


class VerifiedHTTPSConnection(httplib.HTTPSConnection):
        def __init__(self, *args, **kwargs):
            try:
                global socket, ssl
                import socket
                import ssl
            except ImportError:
                raise FatalBackendException(_("Missing socket or ssl python modules."))

            httplib.HTTPSConnection.__init__(self, *args, **kwargs)

            self.cacert_file = globals.ssl_cacert_file
            self.cacert_candidates = ["~/.duplicity/cacert.pem",
                                      "~/duplicity_cacert.pem",
                                      "/etc/duplicity/cacert.pem"]
            # if no cacert file was given search default locations
            if not self.cacert_file:
                for path in self.cacert_candidates:
                    path = os.path.expanduser(path)
                    if (os.path.isfile(path)):
                        self.cacert_file = path
                        break

            # check if file is accessible (libssl errors are not very detailed)
            if self.cacert_file and not os.access(self.cacert_file, os.R_OK):
                raise FatalBackendException(_("Cacert database file '%s' is not readable.") %
                                            self.cacert_file)

        def connect(self):
            # create new socket
            sock = socket.create_connection((self.host, self.port),
                                            self.timeout)
            if self._tunnel_host:
                self.sock = sock
                self.tunnel()

            # python 2.7.9+ supports default system certs now
            if "create_default_context" in dir(ssl):
                context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH,
                                                     cafile=self.cacert_file,
                                                     capath=globals.ssl_cacert_path)
                self.sock = context.wrap_socket(sock, server_hostname=self.host)
            # the legacy way needing a cert file
            else:
                if globals.ssl_cacert_path:
                    raise FatalBackendException(
                        _("Option '--ssl-cacert-path' is not supported "
                          "with python 2.7.8 and below."))

                if not self.cacert_file:
                    raise FatalBackendException(_("""\
For certificate verification with python 2.7.8 or earlier a cacert database
file is needed in one of these locations: %s
Hints:
  Consult the man page, chapter 'SSL Certificate Verification'.
  Consider using the options --ssl-cacert-file, --ssl-no-check-certificate .""") %
                                                ", ".join(self.cacert_candidates))

                # wrap the socket in ssl using verification
                self.sock = ssl.wrap_socket(sock,
                                            cert_reqs=ssl.CERT_REQUIRED,
                                            ca_certs=self.cacert_file,
                                            )

        def request(self, *args, **kwargs):  # pylint: disable=method-hidden
            try:
                return httplib.HTTPSConnection.request(self, *args, **kwargs)
            except ssl.SSLError as e:
                # encapsulate ssl errors
                raise BackendException("SSL failed: %s" % util.uexc(e),
                                       log.ErrorCode.backend_error)


class WebDAVBackend(duplicity.backend.Backend):
    """Backend for accessing a WebDAV repository.

    webdav backend contributed in 2006 by Jesper Zedlitz <jesper@zedlitz.de>
    """

    """
    Request just the names.
    """
    listbody = '<?xml version="1.0"?><D:propfind xmlns:D="DAV:"><D:prop><D:resourcetype/></D:prop></D:propfind>'

    """Connect to remote store using WebDAV Protocol"""
    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)
        self.headers = {'Connection': 'keep-alive'}
        self.parsed_url = parsed_url
        self.digest_challenge = None
        self.digest_auth_handler = None

        self.username = parsed_url.username
        self.password = self.get_password()
        self.directory = self.sanitize_path(parsed_url.path)

        log.Info(_("Using WebDAV protocol %s") % (globals.webdav_proto,))
        log.Info(_("Using WebDAV host %s port %s") % (parsed_url.hostname,
                                                      parsed_url.port))
        log.Info(_("Using WebDAV directory %s") % (self.directory,))

        self.conn = None

    def sanitize_path(self, path):
        if path:
            foldpath = re.compile('/+')
            return foldpath.sub('/', path + '/')
        else:
            return '/'

    def getText(self, nodelist):
        rc = ""
        for node in nodelist:
            if node.nodeType == node.TEXT_NODE:
                rc = rc + node.data
        return rc

    def _retry_cleanup(self):
        self.connect(forced=True)

    def connect(self, forced=False):
        """
        Connect or re-connect to the server, updates self.conn
        # reconnect on errors as a precaution, there are errors e.g.
        # "[Errno 32] Broken pipe" or SSl errors that render the connection unusable
        """
        if not forced and self.conn \
                and self.conn.host == self.parsed_url.hostname:
            return

        log.Info(_("WebDAV create connection on '%s'") % (self.parsed_url.hostname))
        self._close()
        # http schemes needed for redirect urls from servers
        if self.parsed_url.scheme in ['webdav', 'http']:
            self.conn = httplib.HTTPConnection(self.parsed_url.hostname, self.parsed_url.port)
        elif self.parsed_url.scheme in ['webdavs', 'https']:
            if globals.ssl_no_check_certificate:
                self.conn = httplib.HTTPSConnection(self.parsed_url.hostname, self.parsed_url.port)
            else:
                self.conn = VerifiedHTTPSConnection(self.parsed_url.hostname, self.parsed_url.port)
        else:
            raise FatalBackendException(_("WebDAV Unknown URI scheme: %s") % (self.parsed_url.scheme))

    def _close(self):
        if self.conn:
            self.conn.close()

    def request(self, method, path, data=None, redirected=0):
        """
        Wraps the connection.request method to retry once if authentication is
        required
        """
        self._close()  # or we get previous request's data or exception
        self.connect()

        quoted_path = urllib.quote(path, "/:~")

        if self.digest_challenge is not None:
            self.headers['Authorization'] = self.get_digest_authorization(path)

        log.Info(_("WebDAV %s %s request with headers: %s ") % (method, quoted_path, self.headers))
        log.Info(_("WebDAV data length: %s ") % len(str(data)))
        self.conn.request(method, quoted_path, data, self.headers)
        response = self.conn.getresponse()
        log.Info(_("WebDAV response status %s with reason '%s'.") % (response.status, response.reason))
        # resolve redirects and reset url on listing requests (they usually come before everything else)
        if response.status in [301, 302] and method == 'PROPFIND':
            redirect_url = response.getheader('location', None)
            response.close()
            if redirect_url:
                log.Notice(_("WebDAV redirect to: %s ") % urllib.unquote(redirect_url))
                if redirected > 10:
                    raise FatalBackendException(_("WebDAV redirected 10 times. Giving up."))
                self.parsed_url = duplicity.backend.ParsedUrl(redirect_url)
                self.directory = self.sanitize_path(self.parsed_url.path)
                return self.request(method, self.directory, data, redirected + 1)
            else:
                raise FatalBackendException(_("WebDAV missing location header in redirect response."))
        elif response.status == 401:
            response.read()
            response.close()
            self.headers['Authorization'] = self.get_authorization(response, quoted_path)
            log.Info(_("WebDAV retry request with authentification headers."))
            log.Info(_("WebDAV %s %s request2 with headers: %s ") % (method, quoted_path, self.headers))
            log.Info(_("WebDAV data length: %s ") % len(str(data)))
            self.conn.request(method, quoted_path, data, self.headers)
            response = self.conn.getresponse()
            log.Info(_("WebDAV response2 status %s with reason '%s'.") % (response.status, response.reason))

        return response

    def get_authorization(self, response, path):
        """
        Fetches the auth header based on the requested method (basic or digest)
        """
        try:
            auth_hdr = response.getheader('www-authenticate', '')
            token, challenge = auth_hdr.split(' ', 1)
        except ValueError:
            return None
        if token.split(',')[0].lower() == 'negotiate':
            try:
                return self.get_kerberos_authorization()
            except ImportError:
                log.Warn(_("python-kerberos needed to use kerberos \
                          authorization, falling back to basic auth."))
                return self.get_basic_authorization()
            except Exception as e:
                log.Warn(_("Kerberos authorization failed: %s.\
                          Falling back to basic auth.") % e)
                return self.get_basic_authorization()
        elif token.lower() == 'basic':
            return self.get_basic_authorization()
        else:
            self.digest_challenge = self.parse_digest_challenge(challenge)
            return self.get_digest_authorization(path)

    def parse_digest_challenge(self, challenge_string):
        return urllib2.parse_keqv_list(urllib2.parse_http_list(challenge_string))

    def get_kerberos_authorization(self):
        import kerberos  # pylint: disable=import-error
        _, ctx = kerberos.authGSSClientInit("HTTP@%s" % self.conn.host)
        kerberos.authGSSClientStep(ctx, "")
        tgt = kerberos.authGSSClientResponse(ctx)
        return 'Negotiate %s' % tgt

    def get_basic_authorization(self):
        """
        Returns the basic auth header
        """
        auth_string = '%s:%s' % (self.username, self.password)
        return 'Basic %s' % base64.encodestring(auth_string).strip()

    def get_digest_authorization(self, path):
        """
        Returns the digest auth header
        """
        u = self.parsed_url
        if self.digest_auth_handler is None:
            pw_manager = urllib2.HTTPPasswordMgrWithDefaultRealm()
            pw_manager.add_password(None, self.conn.host, self.username, self.password)
            self.digest_auth_handler = urllib2.HTTPDigestAuthHandler(pw_manager)

        # building a dummy request that gets never sent,
        # needed for call to auth_handler.get_authorization
        scheme = u.scheme == 'webdavs' and 'https' or 'http'
        hostname = u.port and "%s:%s" % (u.hostname, u.port) or u.hostname
        dummy_url = "%s://%s%s" % (scheme, hostname, path)
        dummy_req = CustomMethodRequest(self.conn._method, dummy_url)
        auth_string = self.digest_auth_handler.get_authorization(dummy_req,
                                                                 self.digest_challenge)
        return 'Digest %s' % auth_string

    def _list(self):
        response = None
        try:
            self.headers['Depth'] = "1"
            response = self.request("PROPFIND", self.directory, self.listbody)
            del self.headers['Depth']
            # if the target collection does not exist, create it.
            if response.status == 404:
                response.close()  # otherwise next request fails with ResponseNotReady
                self.makedir()
                # just created an empty folder, so return empty
                return []
            elif response.status in [200, 207]:
                document = response.read()
                response.close()
            else:
                status = response.status
                reason = response.reason
                response.close()
                raise BackendException("Bad status code %s reason %s." % (status, reason))

            log.Debug("%s" % (document,))
            dom = xml.dom.minidom.parseString(document)
            result = []
            for href in dom.getElementsByTagName('d:href') + dom.getElementsByTagName('D:href'):
                filename = self.taste_href(href)
                if filename:
                    result.append(filename)
            return result
        except Exception as e:
            raise e
        finally:
            if response:
                response.close()

    def makedir(self):
        """Make (nested) directories on the server."""
        dirs = self.directory.split("/")
        # url causes directory to start with /, but it might be given
        # with or without trailing / (which is required)
        if dirs[-1] == '':
            dirs = dirs[0:-1]
        for i in range(1, len(dirs)):
            d = "/".join(dirs[0:i + 1]) + "/"

            self.headers['Depth'] = "1"
            response = self.request("PROPFIND", d)
            del self.headers['Depth']

            log.Info("Checking existence dir %s: %d" % (d, response.status))

            if response.status == 404:
                log.Info(_("Creating missing directory %s") % d)

                res = self.request("MKCOL", d)
                if res.status != 201:
                    raise BackendException(_("WebDAV MKCOL %s failed: %s %s") %
                                           (d, res.status, res.reason))

    def taste_href(self, href):
        """
        Internal helper to taste the given href node and, if
        it is a duplicity file, collect it as a result file.

        @return: A matching filename, or None if the href did not match.
        """
        raw_filename = self.getText(href.childNodes).strip()
        parsed_url = urlparse.urlparse(urllib.unquote(raw_filename))
        filename = parsed_url.path
        log.Debug(_("WebDAV path decoding and translation: "
                  "%s -> %s") % (raw_filename, filename))

        # at least one WebDAV server returns files in the form
        # of full URL:s. this may or may not be
        # according to the standard, but regardless we
        # feel we want to bail out if the hostname
        # does not match until someone has looked into
        # what the WebDAV protocol mandages.
        if parsed_url.hostname is not None \
           and not (parsed_url.hostname == self.parsed_url.hostname):
            m = "Received filename was in the form of a "\
                "full url, but the hostname (%s) did "\
                "not match that of the webdav backend "\
                "url (%s) - aborting as a conservative "\
                "safety measure. If this happens to you, "\
                "please report the problem"\
                "" % (parsed_url.hostname,
                      self.parsed_url.hostname)
            raise BackendException(m)

        if filename.startswith(self.directory):
            filename = filename.replace(self.directory, '', 1)
            return filename
        else:
            return None

    def _get(self, remote_filename, local_path):
        url = self.directory + remote_filename
        response = None
        try:
            target_file = local_path.open("wb")
            response = self.request("GET", url)
            if response.status == 200:
                # data=response.read()
                target_file.write(response.read())
                # import hashlib
                # log.Info("WebDAV GOT %s bytes with md5=%s" %
                # (len(data),hashlib.md5(data).hexdigest()) )
                assert not target_file.close()
                response.close()
            else:
                status = response.status
                reason = response.reason
                response.close()
                raise BackendException(_("WebDAV GET Bad status code %s reason %s.") %
                                       (status, reason))
        except Exception as e:
            raise e
        finally:
            if response:
                response.close()

    def _put(self, source_path, remote_filename):
        url = self.directory + remote_filename
        response = None
        try:
            source_file = source_path.open("rb")
            response = self.request("PUT", url, source_file.read())
            # 200 is returned if a file is overwritten during restarting
            if response.status in [200, 201, 204]:
                response.read()
                response.close()
            else:
                status = response.status
                reason = response.reason
                response.close()
                raise BackendException(_("WebDAV PUT Bad status code %s reason %s.") %
                                       (status, reason))
        except Exception as e:
            raise e
        finally:
            if response:
                response.close()

    def _delete(self, filename):
        url = self.directory + filename
        response = None
        try:
            response = self.request("DELETE", url)
            if response.status in [200, 204]:
                response.read()
                response.close()
            else:
                status = response.status
                reason = response.reason
                response.close()
                raise BackendException(_("WebDAV DEL Bad status code %s reason %s.") %
                                       (status, reason))
        except Exception as e:
            raise e
        finally:
            if response:
                response.close()

duplicity.backend.register_backend("http", WebDAVBackend)
duplicity.backend.register_backend("https", WebDAVBackend)
duplicity.backend.register_backend("webdav", WebDAVBackend)
duplicity.backend.register_backend("webdavs", WebDAVBackend)
duplicity.backend.uses_netloc.extend(['http', 'https', 'webdav', 'webdavs'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Gu1
# Licensed under the MIT license

import os
import pyrax
import pyrax.exceptions as exc
import requests
import re
import urlparse
import ConfigParser
import time
from pyrax.base_identity import BaseIdentity, Service
from requests.compat import quote, quote_plus


OAUTH_ENDPOINT = "https://api.hubic.com/oauth/"
API_ENDPOINT = "https://api.hubic.com/1.0/"
TOKENS_FILE = os.path.expanduser("~/.hubic_tokens")


class BearerTokenAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, req):
        req.headers['Authorization'] = 'Bearer ' + self.token
        return req


class HubicIdentity(BaseIdentity):

    def _get_auth_endpoint(self):
        return ""

    def set_credentials(self, email, password, client_id,
                        client_secret, redirect_uri,
                        authenticate=False):
        """Sets the username and password directly."""
        self._email = email
        self._password = password
        self._client_id = client_id
        self.tenant_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        if authenticate:
            self.authenticate()

    def _read_credential_file(self, cfg):
        """
        Parses the credential file with Rackspace-specific labels.
        """
        self._email = cfg.get("hubic", "email")
        self._password = cfg.get("hubic", "password")
        self._client_id = cfg.get("hubic", "client_id")
        self.tenant_id = self._client_id
        self._client_secret = cfg.get("hubic", "client_secret")
        self._redirect_uri = cfg.get("hubic", "redirect_uri")

    def _parse_error(self, resp):
        if 'location' not in resp.headers:
            return None
        query = urlparse.urlsplit(resp.headers['location']).query
        qs = dict(urlparse.parse_qsl(query))
        return {'error': qs['error'], 'error_description': qs['error_description']}

    def _get_access_token(self, code):
        r = requests.post(
            OAUTH_ENDPOINT + 'token/',
            data={
                'code': code,
                'redirect_uri': self._redirect_uri,
                'grant_type': 'authorization_code',
            },
            auth=(self._client_id, self._client_secret)
        )
        if r.status_code != 200:
            try:
                err = r.json()
                err['code'] = r.status_code
            except:
                err = {}

            raise exc.AuthenticationFailed("Unable to get oauth access token, "
                                           "wrong client_id or client_secret ? (%s)" %
                                           str(err))

        oauth_token = r.json()

        config = ConfigParser.ConfigParser()
        config.read(TOKENS_FILE)

        if not config.has_section("hubic"):
            config.add_section("hubic")

        if oauth_token['access_token'] is not None:
            config.set("hubic", "access_token", oauth_token['access_token'])
            with open(TOKENS_FILE, 'wb') as configfile:
                config.write(configfile)
        else:
            raise exc.AuthenticationFailed(
                "Unable to get oauth access token, wrong client_id or client_secret ? (%s)" %
                str(err))

        if oauth_token['refresh_token'] is not None:
            config.set("hubic", "refresh_token", oauth_token['refresh_token'])
            with open(TOKENS_FILE, 'wb') as configfile:
                config.write(configfile)
        else:
            raise exc.AuthenticationFailed("Unable to get the refresh token.")

        # removing username and password from .hubic_tokens
        if config.has_option("hubic", "email"):
            config.remove_option("hubic", "email")
            with open(TOKENS_FILE, 'wb') as configfile:
                config.write(configfile)
            print "username has been removed from the .hubic_tokens file sent to the CE."
        if config.has_option("hubic", "password"):
            config.remove_option("hubic", "password")
            with open(TOKENS_FILE, 'wb') as configfile:
                config.write(configfile)
            print "password has been removed from the .hubic_tokens file sent to the CE."

        return oauth_token

    def _refresh_access_token(self):

        config = ConfigParser.ConfigParser()
        config.read(TOKENS_FILE)
        refresh_token = config.get("hubic", "refresh_token")

        if refresh_token is None:
            raise exc.AuthenticationFailed("refresh_token is null. Not acquiered before ?")

        success = False
        max_retries = 20
        retries = 0
        sleep_time = 30
        max_sleep_time = 3600

        while retries < max_retries and not success:
            r = requests.post(
                OAUTH_ENDPOINT + 'token/',
                data={
                    'refresh_token': refresh_token,
                    'grant_type': 'refresh_token',
                },
                auth=(self._client_id, self._client_secret)
            )
            if r.status_code != 200:
                if r.status_code == 509:
                    print "status_code 509: attempt #", retries, " failed"
                    retries += 1
                    time.sleep(sleep_time)
                    sleep_time = sleep_time * 2
                    if sleep_time > max_sleep_time:
                        sleep_time = max_sleep_time
                else:
                    try:
                        err = r.json()
                        err['code'] = r.status_code
                    except:
                        err = {}

                    raise exc.AuthenticationFailed(
                        "Unable to get oauth access token, wrong client_id or client_secret ? (%s)" %
                        str(err))
            else:
                success = True

        if not success:
            raise exc.AuthenticationFailed(
                "All the attempts failed to get the refresh token: "
                "status_code = 509: Bandwidth Limit Exceeded")

        oauth_token = r.json()

        if oauth_token['access_token'] is not None:
            return oauth_token
        else:
            raise exc.AuthenticationFailed("Unable to get oauth access token from json")

    def authenticate(self):
        config = ConfigParser.ConfigParser()
        config.read(TOKENS_FILE)

        if config.has_option("hubic", "refresh_token"):
            oauth_token = self._refresh_access_token()
        else:
            r = requests.get(
                OAUTH_ENDPOINT + 'auth/?client_id={0}&redirect_uri={1}'
                '&scope=credentials.r,account.r&response_type=code&state={2}'.format(
                    quote(self._client_id),
                    quote_plus(self._redirect_uri),
                    pyrax.utils.random_ascii()  # csrf ? wut ?..
                ),
                allow_redirects=False
            )
            if r.status_code != 200:
                raise exc.AuthenticationFailed("Incorrect/unauthorized "
                                               "client_id (%s)" % str(self._parse_error(r)))

            try:
                from lxml import html as lxml_html
            except ImportError:
                lxml_html = None

            if lxml_html:
                oauth = lxml_html.document_fromstring(r.content).xpath('//input[@name="oauth"]')
                oauth = oauth[0].value if oauth else None
            else:
                oauth = re.search(
                    r'<input\s+[^>]*name=[\'"]?oauth[\'"]?\s+[^>]*value=[\'"]?(\d+)[\'"]?>',
                    r.content)
                oauth = oauth.group(1) if oauth else None

            if not oauth:
                raise exc.AuthenticationFailed("Unable to get oauth_id from authorization page")

            if self._email is None or self._password is None:
                raise exc.AuthenticationFailed("Cannot retrieve email and/or password. "
                                               "Please run expresslane-hubic-setup.sh")

            r = requests.post(
                OAUTH_ENDPOINT + 'auth/',
                data={
                    'action': 'accepted',
                    'oauth': oauth,
                    'login': self._email,
                    'user_pwd': self._password,
                    'account': 'r',
                    'credentials': 'r',

                },
                allow_redirects=False
            )

            try:
                query = urlparse.urlsplit(r.headers['location']).query
                code = dict(urlparse.parse_qsl(query))['code']
            except:
                raise exc.AuthenticationFailed("Unable to authorize client_id, "
                                               "invalid login/password ?")

            oauth_token = self._get_access_token(code)

        if oauth_token['token_type'].lower() != 'bearer':
            raise exc.AuthenticationFailed("Unsupported access token type")

        r = requests.get(
            API_ENDPOINT + 'account/credentials',
            auth=BearerTokenAuth(oauth_token['access_token']),
        )

        swift_token = r.json()
        self.authenticated = True
        self.token = swift_token['token']
        self.expires = swift_token['expires']
        self.services['object_store'] = Service(self, {
            'name': 'HubiC',
            'type': 'cloudfiles',
            'endpoints': [
                {'public_url': swift_token['endpoint']}
            ]
        })
        self.username = self.password = None



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2014 Canonical Ltd
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import os
import StringIO
import unittest

import duplicity.backend
from duplicity import log
from duplicity import path
from duplicity.errors import BackendException
from . import UnitTestCase


class BackendInstanceBase(UnitTestCase):

    def setUp(self):
        UnitTestCase.setUp(self)
        assert not os.system("rm -rf testfiles")
        os.makedirs('testfiles')
        self.backend = None
        self.local = path.Path('testfiles/local')
        self.local.writefileobj(StringIO.StringIO("hello"))

    def tearDown(self):
        if self.backend is None:
            return
        if hasattr(self.backend, '_close'):
            self.backend._close()

    def test_get(self):
        if self.backend is None:
            return
        self.backend._put(self.local, 'a')
        getfile = path.Path('testfiles/getfile')
        self.backend._get('a', getfile)
        self.assertTrue(self.local.compare_data(getfile))

    def test_list(self):
        if self.backend is None:
            return
        self.backend._put(self.local, 'a')
        self.backend._put(self.local, 'b')
        # It's OK for backends to create files as a side effect of put (e.g.
        # the par2 backend does), so only check that at least a and b exist.
        self.assertTrue('a' in self.backend._list())
        self.assertTrue('b' in self.backend._list())

    def test_delete(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, '_delete'):
            self.assertTrue(hasattr(self.backend, '_delete_list'))
            return
        self.backend._put(self.local, 'a')
        self.backend._put(self.local, 'b')
        self.backend._delete('a')
        self.assertFalse('a' in self.backend._list())
        self.assertTrue('b' in self.backend._list())

    def test_delete_clean(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, '_delete'):
            self.assertTrue(hasattr(self.backend, '_delete_list'))
            return
        self.backend._put(self.local, 'a')
        self.backend._delete('a')
        self.assertEqual(self.backend._list(), [])

    def test_delete_missing(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, '_delete'):
            self.assertTrue(hasattr(self.backend, '_delete_list'))
            return
        # Backends can either silently ignore this, or throw an error
        # that gives log.ErrorCode.backend_not_found.
        try:
            self.backend._delete('a')
        except BackendException as e:
            pass  # Something went wrong, but it was an 'expected' something
        except Exception as e:
            code = duplicity.backend._get_code_from_exception(self.backend, 'delete', e)
            self.assertEqual(code, log.ErrorCode.backend_not_found)

    def test_delete_list(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, '_delete_list'):
            self.assertTrue(hasattr(self.backend, '_delete'))
            return
        self.backend._put(self.local, 'a')
        self.backend._put(self.local, 'b')
        self.backend._put(self.local, 'c')
        self.backend._delete_list(['a', 'd', 'c'])
        files = self.backend._list()
        self.assertFalse('a' in files, files)
        self.assertTrue('b' in files, files)
        self.assertFalse('c' in files, files)

    def test_move(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, '_move'):
            return

        copy = path.Path('testfiles/copy')
        self.local.copy(copy)

        self.backend._move(self.local, 'a')
        self.assertTrue('a' in self.backend._list())
        self.assertFalse(self.local.exists())

        getfile = path.Path('testfiles/getfile')
        self.backend._get('a', getfile)
        self.assertTrue(copy.compare_data(getfile))

    def test_query_exists(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, '_query'):
            return
        self.backend._put(self.local, 'a')
        info = self.backend._query('a')
        self.assertEqual(info['size'], self.local.getsize())

    def test_query_missing(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, '_query'):
            return
        # Backends can either return -1 themselves, or throw an error
        # that gives log.ErrorCode.backend_not_found.
        try:
            info = self.backend._query('a')
        except BackendException as e:
            pass  # Something went wrong, but it was an 'expected' something
        except Exception as e:
            code = duplicity.backend._get_code_from_exception(self.backend, 'query', e)
            self.assertEqual(code, log.ErrorCode.backend_not_found)
        else:
            self.assertEqual(info['size'], -1)

    def test_query_list(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, '_query_list'):
            return
        self.backend._put(self.local, 'a')
        self.backend._put(self.local, 'c')
        info = self.backend._query_list(['a', 'b'])
        self.assertEqual(info['a']['size'], self.local.getsize())
        self.assertEqual(info['b']['size'], -1)
        self.assertFalse('c' in info)


class LocalBackendTest(BackendInstanceBase):
    def setUp(self):
        super(LocalBackendTest, self).setUp()
        url = 'file://testfiles/output'
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, 'LocalBackend')


class Par2BackendTest(BackendInstanceBase):
    def setUp(self):
        super(Par2BackendTest, self).setUp()
        url = 'par2+file://testfiles/output'
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, 'Par2Backend')

    # TODO: Add par2-specific tests here, to confirm that we can recover from
    # a missing file


# class RsyncBackendTest(BackendInstanceBase):
#     def setUp(self):
#         super(RsyncBackendTest, self).setUp()
#         os.makedirs('testfiles/output')  # rsync needs it to exist first
#         url = 'rsync://%s/testfiles/output' % os.getcwd()
#         self.backend = duplicity.backend.get_backend_object(url)
#         self.assertEqual(self.backend.__class__.__name__, 'RsyncBackend')


class TahoeBackendTest(BackendInstanceBase):
    def setUp(self):
        super(TahoeBackendTest, self).setUp()
        os.makedirs('testfiles/output')
        url = 'tahoe://testfiles/output'
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, 'TAHOEBackend')


class HSIBackendTest(BackendInstanceBase):
    def setUp(self):
        super(HSIBackendTest, self).setUp()
        os.makedirs('testfiles/output')
        # hostname is ignored...  Seemingly on purpose
        url = 'hsi://hostname%s/testfiles/output' % os.getcwd()
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, 'HSIBackend')


class FTPBackendTest(BackendInstanceBase):
    def setUp(self):
        super(FTPBackendTest, self).setUp()
        os.makedirs('testfiles/output')
        url = 'ftp://user:pass@hostname/testfiles/output'
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, 'LFTPBackend')


class FTPSBackendTest(BackendInstanceBase):
    def setUp(self):
        super(FTPSBackendTest, self).setUp()
        os.makedirs('testfiles/output')
        url = 'ftps://user:pass@hostname/testfiles/output'
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, 'LFTPBackend')


if __name__ == "__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2015 Stefan Breunig <stefan-duplicity@breunig.xyz>
# Copyright 2015 Malay Shah <malays@gmail.com>
# Based on the backend ncftpbackend.py
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import os.path
import urllib
import string
import json

import duplicity.backend
from duplicity import globals
from duplicity import log
from duplicity.errors import *  # @UnusedWildImport
from duplicity import tempdir


class ACDBackend(duplicity.backend.Backend):
    acd_cmd = 'acd_cli'
    """Connect to remote store using acd_cli"""
    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        # we expect an error return, so go low-level and ignore it
        try:
            p = os.popen(self.acd_cmd + " version")
            fout = p.read()
            ret = p.close()
        except Exception:
            pass
        # the expected error is 0
        if ret is not None:
            log.FatalError(self.acd_cmd + " not found:  Please install acd_cli",
                           log.ErrorCode.backend_not_found)

        self.parsed_url = parsed_url
        self.url_string = duplicity.backend.strip_auth_from_url(self.parsed_url)

        # Use an explicit directory name.
        if self.url_string[-1] != '/':
            self.url_string += '/'

        self.subprocess_popen(self.acd_cmd + " sync")

        # Initialize remote directory
        remote_path = urllib.unquote(self.parsed_url.path.replace('///', '/')).rstrip()
        commandline = self.acd_cmd + " ls '%s'" % (remote_path)
        try:
            self.subprocess_popen(commandline)
        except BackendException as e:
            if e.code == 50:  # Remote directory not there, create a new one
                commandline = self.acd_cmd + " mkdir '%s'" % remote_path
                self.subprocess_popen(commandline)

    def _put(self, source_path, remote_filename=None):
        """Transfer source_path to remote_filename"""
        if not remote_filename:
            remote_filename = source_path.get_filename()

        # WORKAROUND for acd_cli: cannot specify remote filename
        # Link tmp file to the desired remote filename locally and upload
        remote_path = urllib.unquote(self.parsed_url.path.replace('///', '/'))
        local_real_duplicity_file = os.path.join(os.path.dirname(source_path.name), remote_filename.rstrip())

        deleteFile = False
        if(source_path.name != local_real_duplicity_file):
            try:
                os.symlink(source_path.name, local_real_duplicity_file)
                deleteFile = True
            except IOError as e:
                log.FatalError("Unable to copy " + source_path.name + " to " + local_real_duplicity_file)

        commandline = self.acd_cmd + " upload --force --overwrite '%s' '%s'" % \
            (local_real_duplicity_file, remote_path)

        try:
            l = self.subprocess_popen(commandline)
        finally:
            if (deleteFile):
                try:
                    os.remove(local_real_duplicity_file)
                except OSError as e:
                    log.FatalError("Unable to remove file %s" % e)

    def _get(self, remote_filename, local_path):
        """Get remote filename, saving it to local_path"""
        remote_path = os.path.join(urllib.unquote(self.parsed_url.path.replace('///', '/')), remote_filename).rstrip()
        local_dir = os.path.dirname(local_path.name)
        local_filename = os.path.basename(local_path.name)
        commandline = self.acd_cmd + " download '%s' '%s'" % \
            (remote_path, local_dir)
        l = self.subprocess_popen(commandline)

        # Keep the remote filename and move the file over
        try:
            os.rename(os.path.join(local_dir, remote_filename), local_path.name)
        except IOError as e:
            log.FatalError("Unable to move file %s" % e)

        local_path.setdata()
        if not local_path.exists():
            raise BackendException("File %s not found" % local_path.name)

    def _query(self, remote_filename):
        remote_path = os.path.join(urllib.unquote(self.parsed_url.path.replace('///', '/')), remote_filename).rstrip()
        commandline = self.acd_cmd + " metadata '%s'" % remote_path
        node = self.subprocess_popen(commandline)
        if (node[0] == 0 and node[1]):
            try:
                size = json.loads(node[1])['contentProperties']['size']
                return {'size': size}
            except ValueError as e:
                raise BackendException('Malformed JSON: expected "contentProperties->size" member in %s' % node[1])
        return {'size': -1}

    def _list(self):
        """List files in directory"""
        def dir_split(str):
            if (str):
                return str.split()[2]
            else:
                return None
        commandline = self.acd_cmd + " ls '%s'" % self.parsed_url.path.replace('///', '/')
        l = self.subprocess_popen(commandline)
        return [x for x in map(dir_split, l[1].split('\n')) if x]

    def _delete(self, remote_filename):
        """Delete remote_filename"""
        remote_file_path = os.path.join(urllib.unquote(self.parsed_url.path.replace('///', '/')),
                                        remote_filename).rstrip()
        commandline = self.acd_cmd + " rm '%s'" % (remote_file_path)
        self.subprocess_popen(commandline)

duplicity.backend.register_backend("acd+acdcli", ACDBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2013 Matthieu Huin <mhu@enovance.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import os

import duplicity.backend
from duplicity import log
from duplicity import util
from duplicity.errors import BackendException


class SwiftBackend(duplicity.backend.Backend):
    """
    Backend for Swift
    """
    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        try:
            from swiftclient import Connection
            from swiftclient import ClientException
        except ImportError as e:
            raise BackendException("""\
Swift backend requires the python-swiftclient library.
Exception: %s""" % str(e))

        self.resp_exc = ClientException
        conn_kwargs = {}

        # if the user has already authenticated
        if 'SWIFT_PREAUTHURL' in os.environ and 'SWIFT_PREAUTHTOKEN' in os.environ:
            conn_kwargs['preauthurl'] = os.environ['SWIFT_PREAUTHURL']
            conn_kwargs['preauthtoken'] = os.environ['SWIFT_PREAUTHTOKEN']

        else:
            if 'SWIFT_USERNAME' not in os.environ:
                raise BackendException('SWIFT_USERNAME environment variable '
                                       'not set.')

            if 'SWIFT_PASSWORD' not in os.environ:
                raise BackendException('SWIFT_PASSWORD environment variable '
                                       'not set.')

            if 'SWIFT_AUTHURL' not in os.environ:
                raise BackendException('SWIFT_AUTHURL environment variable '
                                       'not set.')

            conn_kwargs['user'] = os.environ['SWIFT_USERNAME']
            conn_kwargs['key'] = os.environ['SWIFT_PASSWORD']
            conn_kwargs['authurl'] = os.environ['SWIFT_AUTHURL']

        os_options = {}

        if 'SWIFT_AUTHVERSION' in os.environ:
            conn_kwargs['auth_version'] = os.environ['SWIFT_AUTHVERSION']
            if os.environ['SWIFT_AUTHVERSION'] == '3':
                if 'SWIFT_USER_DOMAIN_NAME' in os.environ:
                    os_options.update({'user_domain_name': os.environ['SWIFT_USER_DOMAIN_NAME']})
                if 'SWIFT_USER_DOMAIN_ID' in os.environ:
                    os_options.update({'user_domain_id': os.environ['SWIFT_USER_DOMAIN_ID']})
                if 'SWIFT_PROJECT_DOMAIN_NAME' in os.environ:
                    os_options.update({'project_domain_name': os.environ['SWIFT_PROJECT_DOMAIN_NAME']})
                if 'SWIFT_PROJECT_DOMAIN_ID' in os.environ:
                    os_options.update({'project_domain_id': os.environ['SWIFT_PROJECT_DOMAIN_ID']})
                if 'SWIFT_TENANTNAME' in os.environ:
                    os_options.update({'tenant_name': os.environ['SWIFT_TENANTNAME']})
                if 'SWIFT_ENDPOINT_TYPE' in os.environ:
                    os_options.update({'endpoint_type': os.environ['SWIFT_ENDPOINT_TYPE']})
                if 'SWIFT_USERID' in os.environ:
                    os_options.update({'user_id': os.environ['SWIFT_USERID']})
                if 'SWIFT_TENANTID' in os.environ:
                    os_options.update({'tenant_id': os.environ['SWIFT_TENANTID']})
                if 'SWIFT_REGIONNAME' in os.environ:
                    os_options.update({'region_name': os.environ['SWIFT_REGIONNAME']})

        else:
            conn_kwargs['auth_version'] = '1'
        if 'SWIFT_TENANTNAME' in os.environ:
            conn_kwargs['tenant_name'] = os.environ['SWIFT_TENANTNAME']
        if 'SWIFT_REGIONNAME' in os.environ:
            os_options.update({'region_name': os.environ['SWIFT_REGIONNAME']})

        conn_kwargs['os_options'] = os_options

        # This folds the null prefix and all null parts, which means that:
        #  //MyContainer/ and //MyContainer are equivalent.
        #  //MyContainer//My/Prefix/ and //MyContainer/My/Prefix are equivalent.
        url_parts = [x for x in parsed_url.path.split('/') if x != '']

        self.container = url_parts.pop(0)
        if url_parts:
            self.prefix = '%s/' % '/'.join(url_parts)
        else:
            self.prefix = ''

        container_metadata = None
        try:
            self.conn = Connection(**conn_kwargs)
            container_metadata = self.conn.head_container(self.container)
        except ClientException:
            pass
        except Exception as e:
            log.FatalError("Connection failed: %s %s"
                           % (e.__class__.__name__, str(e)),
                           log.ErrorCode.connection_failed)

        if container_metadata is None:
            log.Info("Creating container %s" % self.container)
            try:
                self.conn.put_container(self.container)
            except Exception as e:
                log.FatalError("Container creation failed: %s %s"
                               % (e.__class__.__name__, str(e)),
                               log.ErrorCode.connection_failed)

    def _error_code(self, operation, e):
        if isinstance(e, self.resp_exc):
            if e.http_status == 404:
                return log.ErrorCode.backend_not_found

    def _put(self, source_path, remote_filename):
        self.conn.put_object(self.container, self.prefix + remote_filename,
                             file(source_path.name))

    def _get(self, remote_filename, local_path):
        headers, body = self.conn.get_object(self.container, self.prefix + remote_filename, resp_chunk_size=1024)
        with open(local_path.name, 'wb') as f:
            for chunk in body:
                f.write(chunk)

    def _list(self):
        headers, objs = self.conn.get_container(self.container, full_listing=True, path=self.prefix)
        # removes prefix from return values. should check for the prefix ?
        return [o['name'][len(self.prefix):] for o in objs]

    def _delete(self, filename):
        self.conn.delete_object(self.container, self.prefix + filename)

    def _query(self, filename):
        sobject = self.conn.head_object(self.container, self.prefix + filename)
        return {'size': int(sobject['content-length'])}


duplicity.backend.register_backend("swift", SwiftBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2008 Michael Terry <mike@mterry.name>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import unittest
import os

from . import FunctionalTestCase


class LogTest(FunctionalTestCase):
    """Test machine-readable functions/classes in log.py"""

    def setUp(self):
        super(LogTest, self).setUp()
        assert not os.system("rm -f /tmp/duplicity.log")

    def tearDown(self):
        assert not os.system("rm -f /tmp/duplicity.log")
        super(LogTest, self).tearDown()

    def test_command_line_error(self):
        """Check notification of a simple error code"""

        # Run actual duplicity command (will fail, because no arguments passed)
        os.system("duplicity --log-file=/tmp/duplicity.log >/dev/null 2>&1")

        # The format of the file should be:
        # """ERROR 2
        # . Blah blah blah.
        # . Blah blah blah.
        #
        # """
        f = open('/tmp/duplicity.log', 'r')
        linecount = 0
        lastline = False
        for line in f:
            assert(not lastline)
            linecount += 1
            if linecount == 1:
                assert(line == "ERROR 2\n")
            elif line != "\n":
                assert(line.startswith(". "))
            else:
                lastline = True
        assert(lastline)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python2
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import sys
import os
from setuptools import setup, Extension
from setuptools.command.test import test
from setuptools.command.install import install
from setuptools.command.sdist import sdist
from distutils.command.build_scripts import build_scripts

version_string = "0.7.17"

if sys.version_info[:2] < (2, 6) or sys.version_info[:2] > (2, 7):
    print("Sorry, duplicity requires version 2.6 or 2.7 of python.")
    sys.exit(1)

incdir_list = libdir_list = None

if os.name == 'posix':
    LIBRSYNC_DIR = os.environ.get('LIBRSYNC_DIR', '')
    args = sys.argv[:]
    for arg in args:
        if arg.startswith('--librsync-dir='):
            LIBRSYNC_DIR = arg.split('=')[1]
            sys.argv.remove(arg)
    if LIBRSYNC_DIR:
        incdir_list = [os.path.join(LIBRSYNC_DIR, 'include')]
        libdir_list = [os.path.join(LIBRSYNC_DIR, 'lib')]

data_files = [('share/man/man1',
               ['bin/duplicity.1',
                'bin/rdiffdir.1']),
              ('share/doc/duplicity-%s' % version_string,
               ['COPYING',
                'README',
                'README-REPO',
                'README-LOG',
                'CHANGELOG']),
              ]

top_dir = os.path.dirname(os.path.abspath(__file__))
assert os.path.exists(os.path.join(top_dir, "po")), "Missing 'po' directory."
for root, dirs, files in os.walk(os.path.join(top_dir, "po")):
    for file in files:
        path = os.path.join(root, file)
        if path.endswith("duplicity.mo"):
            lang = os.path.split(root)[-1]
            data_files.append(
                ('share/locale/%s/LC_MESSAGES' % lang,
                 ["po/%s/duplicity.mo" % lang]))


class TestCommand(test):

    def run(self):
        # Make sure all modules are ready
        build_cmd = self.get_finalized_command("build_py")
        build_cmd.run()
        # And make sure our scripts are ready
        build_scripts_cmd = self.get_finalized_command("build_scripts")
        build_scripts_cmd.run()

        # make symlinks for test data
        if build_cmd.build_lib != top_dir:
            for path in ['testfiles.tar.gz', 'gnupg']:
                src = os.path.join(top_dir, 'testing', path)
                target = os.path.join(build_cmd.build_lib, 'testing', path)
                try:
                    os.symlink(src, target)
                except Exception:
                    pass

        os.environ['PATH'] = "%s:%s" % (
            os.path.abspath(build_scripts_cmd.build_dir),
            os.environ.get('PATH'))

        test.run(self)


class InstallCommand(install):

    def run(self):
        # Normally, install will call build().  But we want to delete the
        # testing dir between building and installing.  So we manually build
        # and mark ourselves to skip building when we run() for real.
        self.run_command('build')
        self.skip_build = True

        # This should always be true, but just to make sure!
        if self.build_lib != top_dir:
            testing_dir = os.path.join(self.build_lib, 'testing')
            os.system("rm -rf %s" % testing_dir)

        install.run(self)


# TODO: move logic from dist/makedist inline
class SDistCommand(sdist):

    def run(self):
        version = version_string
        if version[0] == '$':
            version = "0.0dev"
        os.system(os.path.join(top_dir, "dist", "makedist") + " " + version)
        os.system("mkdir -p " + self.dist_dir)
        os.system("mv duplicity-" + version + ".tar.gz " + self.dist_dir)


# don't touch my shebang
class BSCommand (build_scripts):

    def run(self):
        """
        Copy, chmod each script listed in 'self.scripts'
        essentially this is the stripped
         distutils.command.build_scripts.copy_scripts()
        routine
        """
        from stat import ST_MODE
        from distutils.dep_util import newer
        from distutils import log

        self.mkpath(self.build_dir)
        outfiles = []
        for script in self.scripts:
            outfile = os.path.join(self.build_dir, os.path.basename(script))
            outfiles.append(outfile)

            if not self.force and not newer(script, outfile):
                log.debug("not copying %s (up-to-date)", script)
                continue

            log.info("copying and NOT adjusting %s -> %s", script,
                     self.build_dir)
            self.copy_file(script, outfile)

        if os.name == 'posix':
            for file in outfiles:
                if self.dry_run:
                    log.info("changing mode of %s", file)
                else:
                    oldmode = os.stat(file)[ST_MODE] & 0o7777
                    newmode = (oldmode | 0o555) & 0o7777
                    if newmode != oldmode:
                        log.info("changing mode of %s from %o to %o",
                                 file, oldmode, newmode)
                        os.chmod(file, newmode)


setup(name="duplicity",
      version=version_string,
      description="Encrypted backup using rsync algorithm",
      author="Ben Escoto <ben@emerose.org>",
      author_email="bescoto@stanford.edu",
      maintainer="Kenneth Loafman <kenneth@loafman.com>",
      maintainer_email="kenneth@loafman.com",
      url="http://duplicity.nongnu.org/index.html",
      packages=['duplicity',
                'duplicity.backends',
                'duplicity.backends.pyrax_identity',
                'testing',
                'testing.functional',
                'testing.overrides',
                'testing.unit'],
      package_dir={"duplicity": "duplicity",
                   "duplicity.backends": "duplicity/backends", },
      ext_modules=[Extension("duplicity._librsync",
                             ["duplicity/_librsyncmodule.c"],
                             include_dirs=incdir_list,
                             library_dirs=libdir_list,
                             libraries=["rsync"])],
      scripts=['bin/rdiffdir', 'bin/duplicity'],
      data_files=data_files,
      install_requires=['fasteners'],
      tests_require=['fasteners', 'mock', 'pexpect'],
      test_suite='testing',
      cmdclass={'test': TestCommand,
                'install': InstallCommand,
                'sdist': SDistCommand,
                'build_scripts': BSCommand},
      )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2014 Aaron Whitehouse <aaron@whitehouse.kiwi.nz>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import os
import sys
import platform

if sys.version_info < (2, 7):
    import unittest2 as unittest  # @UnresolvedImport @UnusedImport
else:
    import unittest  # @Reimport

from . import FunctionalTestCase


class IncludeExcludeFunctionalTest(FunctionalTestCase):
    """
    This contains methods used in the tests below for testing the include, exclude and various filelist features.
    """

    # These tests assume the following files and logic, with:
    # "is" meaning that the file is included specifically
    # "ia" meaning that the file should be included automatically because its parent is included
    # "ic" meaning that the folder is included because its contents are included
    # "es" meaning that the file is excluded specifically
    # "ea" meaning that the file should be excluded automatically because its parent is included
    # select2 (es)
    # --- 1.doc (ea)
    # --- 1.py (is)
    # --- 1 (is)
    # ------ 1sub1 (ia)
    # --------- 1sub1sub1 (ia)
    # ------------ 1sub1sub1_file.txt (ia)
    # --------- 1sub1sub2 (es)
    # ------------ 1sub1sub2_file.txt (ea)
    # --------- 1sub1sub3 (ia)
    # ------------ 1sub1sub3_file.txt (es)
    # ------ 1sub2 (ic)
    # --------- 1sub2sub1 (is)
    # --------- 1sub2sub2 (ea)
    # --------- 1sub2sub3 (es)  # Not necessary as also ea, but to ensure there are no issues doing so
    # ------ 1sub3 (ia)
    # --------- 1sub3sub1 (es)
    # --------- 1sub3sub2 (es)
    # --------- 1sub3sub3 (ia)
    # --- 2 (ic)
    # ------ 2sub1 (is)
    # --------- 2sub1sub1 (ia)
    # ------------ 2sub1sub1_file.txt (ia)
    # --------- 2sub1sub2 (es)
    # --------- 2sub1sub3 (es)
    # ------ 2sub2 (ea)
    # --------- 2sub2sub1 (ea)
    # --------- 2sub2sub2 (ea)
    # --------- 2sub2sub3 (ea)
    # ------ 2sub3 (ea)
    # --------- 2sub3sub1 (ea)
    # --------- 2sub3sub3 (ea)
    # --------- 2sub3sub2 (ea)
    # --- 3 (is)
    # ------ 3sub1 (es)
    # --------- 3sub1sub1 (ea)
    # --------- 3sub1sub2 (ea)
    # --------- 3sub1sub3 (ea)
    # ------ 3sub2 (ia)
    # --------- 3sub2sub1 (ia)
    # --------- 3sub2sub2 (ia)
    # --------- 3sub2sub3 (ia)
    # ------ 3sub3 (is)  # Not necessary as also ia, but to ensure there are no issues doing so
    # --------- 3sub3sub1 (ia)
    # --------- 3sub3sub2 (es, ic)
    # ------------ 3sub3sub2_file.txt (is)
    # --------- 3sub3sub3 (ia)
    # --- trailing_space  (ea)  # Note this is "trailing_space ". Excluded until trailing_space test, when (is)
    # ------ trailing_space sub1 (ea)  # Excluded until trailing_space test, when (ia)
    # ------ trailing_space sub2 (ea)  # Excluded until trailing_space test, when (es, ic)
    # --------- trailing_space sub2_file.txt (ea)  # Excluded until trailing_space test, when (is)

    complete_directory_tree = [
        ['1', '2', '3', 'trailing_space ', '1.doc', '1.py'],
        ['1sub1', '1sub2', '1sub3'],
        ['1sub1sub1', '1sub1sub2', '1sub1sub3'],
        ['1sub1sub1_file.txt'],
        ['1sub1sub2_file.txt'],
        ['1sub1sub3_file.txt'],
        ['1sub2sub1', '1sub2sub2', '1sub2sub3'],
        ['1sub3sub1', '1sub3sub2', '1sub3sub3'],
        ['2sub1', '2sub2', '2sub3'],
        ['2sub1sub1', '2sub1sub2', '2sub1sub3'],
        ['2sub1sub1_file.txt'],
        ['2sub2sub1', '2sub2sub2', '2sub2sub3'],
        ['2sub3sub1', '2sub3sub2', '2sub3sub3'],
        ['3sub1', '3sub2', '3sub3'],
        ['3sub1sub1', '3sub1sub2', '3sub1sub3'],
        ['3sub2sub1', '3sub2sub2', '3sub2sub3'],
        ['3sub3sub1', '3sub3sub2', '3sub3sub3'],
        ['3sub3sub2_file.txt'],
        ['trailing_space sub1', 'trailing_space sub2'],
        ['trailing_space sub2_file.txt']
    ]

    expected_restored_tree = [['1', '2', '3', '1.py'],
                              ['1sub1', '1sub2', '1sub3'],
                              ['1sub1sub1', '1sub1sub3'],
                              ['1sub1sub1_file.txt'],
                              ['1sub2sub1'],
                              ['1sub3sub3'],
                              ['2sub1'],
                              ['2sub1sub1'],
                              ['2sub1sub1_file.txt'],
                              ['3sub2', '3sub3'],
                              ['3sub2sub1', '3sub2sub2', '3sub2sub3'],
                              ['3sub3sub1', '3sub3sub2', '3sub3sub3'],
                              ['3sub3sub2_file.txt']]

    expected_restored_tree_with_trailing_space = [['1', '2', '3', 'trailing_space ', '1.py'],
                                                  ['1sub1', '1sub2', '1sub3'],
                                                  ['1sub1sub1', '1sub1sub3'],
                                                  ['1sub1sub1_file.txt'],
                                                  ['1sub2sub1'],
                                                  ['1sub3sub3'],
                                                  ['2sub1'],
                                                  ['2sub1sub1'],
                                                  ['2sub1sub1_file.txt'],
                                                  ['3sub2', '3sub3'],
                                                  ['3sub2sub1', '3sub2sub2', '3sub2sub3'],
                                                  ['3sub3sub1', '3sub3sub2', '3sub3sub3'],
                                                  ['3sub3sub2_file.txt'],
                                                  ['trailing_space sub1', 'trailing_space sub2'],
                                                  ['trailing_space sub2_file.txt']]

    def directory_tree_to_list_of_lists(self, parent_directory):
        """
        This takes a folder as an input and returns a list with its contents. If the directory has subdirectories, it
        returns a list of lists with the contents of those subdirectories.
        """
        directory_list = []
        for root, dirs, files in os.walk(parent_directory):
            to_add = []
            if dirs:
                dirs.sort()  # So that we can easily compare to what we expect
                to_add = dirs
            if files:
                files.sort()  # So that we can easily compare to what we expect
                to_add += files
            if to_add:
                directory_list.append(to_add)
        return directory_list


class TestCheckTestFiles(IncludeExcludeFunctionalTest):
    """ Tests the testfiles required by the exclude/include tests are as expected. """

    def test_files_are_as_expected(self):
        """Test that the contents of testfiles/select are as expected."""
        testfiles = self.directory_tree_to_list_of_lists('testfiles/select2')
        # print(testfiles)
        self.assertEqual(testfiles, self.complete_directory_tree)


class TestIncludeExcludeOptions(IncludeExcludeFunctionalTest):
    """ This tests the behaviour of the duplicity binary when the include/exclude options are passed directly """

    def test_include_exclude_basic(self):
        """ Test --include and --exclude work in the basic case """
        self.backup("full", "testfiles/select2",
                    options=["--include", "testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt",
                             "--exclude", "testfiles/select2/3/3sub3/3sub3sub2",
                             "--include", "testfiles/select2/3/3sub2/3sub2sub2",
                             "--include", "testfiles/select2/3/3sub3",
                             "--exclude", "testfiles/select2/3/3sub1",
                             "--exclude", "testfiles/select2/2/2sub1/2sub1sub3",
                             "--exclude", "testfiles/select2/2/2sub1/2sub1sub2",
                             "--include", "testfiles/select2/2/2sub1",
                             "--exclude", "testfiles/select2/1/1sub3/1sub3sub2",
                             "--exclude", "testfiles/select2/1/1sub3/1sub3sub1",
                             "--exclude", "testfiles/select2/1/1sub2/1sub2sub3",
                             "--include", "testfiles/select2/1/1sub2/1sub2sub1",
                             "--exclude", "testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt",
                             "--exclude", "testfiles/select2/1/1sub1/1sub1sub2",
                             "--exclude", "testfiles/select2/1/1sub2",
                             "--include", "testfiles/select2/1.py",
                             "--include", "testfiles/select2/3",
                             "--include", "testfiles/select2/1",
                             "--exclude", "testfiles/select2/**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_include_exclude_trailing_whitespace(self):
        """Test that folders with trailing whitespace in the names work correctly when passing as include/exclude"""
        # Note that, because this only passes items in as a list of options, this test does not test whether duplicity
        # would correctly interpret commandline options with spaces. However, bin/duplicity uses sys.argv[1:], which
        # should return a list of strings after having correctly processed quotes etc.
        self.backup("full", "testfiles/select2",
                    options=["--include",
                             "testfiles/select2/trailing_space /trailing_space sub2/trailing_space sub2_file.txt",
                             "--exclude", "testfiles/select2/trailing_space /trailing_space sub2",
                             "--include", "testfiles/select2/trailing_space ",
                             "--include", "testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt",
                             "--exclude", "testfiles/select2/3/3sub3/3sub3sub2",
                             "--include", "testfiles/select2/3/3sub2/3sub2sub2",
                             "--include", "testfiles/select2/3/3sub3",
                             "--exclude", "testfiles/select2/3/3sub1",
                             "--exclude", "testfiles/select2/2/2sub1/2sub1sub3",
                             "--exclude", "testfiles/select2/2/2sub1/2sub1sub2",
                             "--include", "testfiles/select2/2/2sub1",
                             "--exclude", "testfiles/select2/1/1sub3/1sub3sub2",
                             "--exclude", "testfiles/select2/1/1sub3/1sub3sub1",
                             "--exclude", "testfiles/select2/1/1sub2/1sub2sub3",
                             "--include", "testfiles/select2/1/1sub2/1sub2sub1",
                             "--exclude", "testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt",
                             "--exclude", "testfiles/select2/1/1sub1/1sub1sub2",
                             "--exclude", "testfiles/select2/1/1sub2",
                             "--include", "testfiles/select2/1.py",
                             "--include", "testfiles/select2/3",
                             "--include", "testfiles/select2/1",
                             "--exclude", "testfiles/select2/**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree_with_trailing_space)


class TestExcludeFilelistTest(IncludeExcludeFunctionalTest):
    """
    Test --exclude-filelist using duplicity binary.
    """

    def test_exclude_filelist(self):
        """Test that exclude filelist works in the basic case """
        # As this is an exclude filelist any lines with no +/- modifier should be treated as if they have a -.
        # Create a filelist
        with open('testfiles/exclude.txt', 'w') as f:
            f.write('+ testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n'
                    'testfiles/select2/3/3sub3/3sub3sub2\n'
                    '+ testfiles/select2/3/3sub2/3sub2sub2\n'
                    '+ testfiles/select2/3/3sub3\n'
                    '- testfiles/select2/3/3sub1\n'  # - added to ensure it makes no difference
                    'testfiles/select2/2/2sub1/2sub1sub3\n'
                    'testfiles/select2/2/2sub1/2sub1sub2\n'
                    '+ testfiles/select2/2/2sub1\n'
                    'testfiles/select2/1/1sub3/1sub3sub2\n'
                    'testfiles/select2/1/1sub3/1sub3sub1\n'
                    'testfiles/select2/1/1sub2/1sub2sub3\n'
                    '+ testfiles/select2/1/1sub2/1sub2sub1\n'
                    'testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n'
                    'testfiles/select2/1/1sub1/1sub1sub2\n'
                    '- testfiles/select2/1/1sub2\n'  # - added to ensure it makes no difference
                    '+ testfiles/select2/1.py\n'
                    '+ testfiles/select2/3\n'
                    '+ testfiles/select2/1\n'
                    'testfiles/select2/**')
        self.backup("full", "testfiles/select2", options=["--exclude-filelist=testfiles/exclude.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_exclude_filelist_combined_imperfections(self):
        """Test that exclude filelist works with imperfections in the input file"""
        # This is a combined test for speed reasons. The individual imperfections are tested as unittests in
        # unit/test_selection.
        # Imperfections tested are;
        # * Leading space/spaces before the modifier
        # * Trailing space/spaces after the filename (but before the newline)
        # * Blank lines (newline character only)
        # * Line only containing spaces
        # * Full-line comments with # as the first character and with leading/trailing spaces
        # * Unnecessarily quoted filenames with/without modifier (both " and ')

        # Create a filelist
        with open('testfiles/exclude.txt', 'w') as f:
            f.write('+ testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n'
                    'testfiles/select2/3/3sub3/3sub3sub2\n'
                    '+ testfiles/select2/3/3sub2/3sub2sub2\n'
                    ' + testfiles/select2/3/3sub3\n'  # Note leading space added here
                    '- testfiles/select2/3/3sub1\n'
                    '  testfiles/select2/2/2sub1/2sub1sub3\n'  # Note leading spaces added here
                    '\n'
                    'testfiles/select2/2/2sub1/2sub1sub2\n'
                    ' + testfiles/select2/2/2sub1 \n'  # Note added trailing/leading space here
                    '- "testfiles/select2/1/1sub3/1sub3sub2"\n'  # Unnecessary quotes
                    '# Testing a full-line comment\n'
                    "'testfiles/select2/1/1sub3/1sub3sub1'  \n"  # Note added spaces and quotes here
                    'testfiles/select2/1/1sub2/1sub2sub3\n'
                    '    \n'
                    '+ testfiles/select2/1/1sub2/1sub2sub1\n'
                    '- testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n'
                    'testfiles/select2/1/1sub1/1sub1sub2\n'
                    '     # Testing a full-line comment with leading and trailing spaces     \n'
                    'testfiles/select2/1/1sub2  \n'  # Note added spaces here
                    '+ testfiles/select2/1.py\n'
                    '+ testfiles/select2/3 \n'  # Note added space here
                    '+ testfiles/select2/1\n'
                    '- testfiles/select2/**')
        self.backup("full", "testfiles/select2", options=["--exclude-filelist=testfiles/exclude.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_exclude_globbing_filelist_combined_imperfections(self):
        """Test that exclude globbing filelist works with imperfections in the input file"""
        # Identical to test_exclude_filelist_combined_imperfections and included to ensure that
        # the deprecated --exclude-globbing-filelist function works as expected until it is deliberately removed.
        # This is a combined test for speed reasons. The individual imperfections are tested as unittests in
        # unit/test_selection.
        # Imperfections tested are;
        # * Leading space/spaces before the modifier
        # * Trailing space/spaces after the filename (but before the newline)
        # * Blank lines (newline character only)
        # * Line only containing spaces
        # * Full-line comments with # as the first character and with leading/trailing spaces
        # * Unnecessarily quoted filenames with/without modifier (both " and ')

        # Create a filelist
        with open('testfiles/exclude.txt', 'w') as f:
            f.write('+ testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n'
                    'testfiles/select2/3/3sub3/3sub3sub2\n'
                    '+ testfiles/select2/3/3sub2/3sub2sub2\n'
                    ' + testfiles/select2/3/3sub3\n'  # Note leading space added here
                    '- testfiles/select2/3/3sub1\n'
                    '  testfiles/select2/2/2sub1/2sub1sub3\n'  # Note leading spaces added here
                    '\n'
                    'testfiles/select2/2/2sub1/2sub1sub2\n'
                    ' + testfiles/select2/2/2sub1 \n'  # Note added trailing/leading space here
                    '- "testfiles/select2/1/1sub3/1sub3sub2"\n'  # Unnecessary quotes
                    '# Testing a full-line comment\n'
                    "'testfiles/select2/1/1sub3/1sub3sub1'  \n"  # Note added spaces and quotes here
                    'testfiles/select2/1/1sub2/1sub2sub3\n'
                    '    \n'
                    '+ testfiles/select2/1/1sub2/1sub2sub1\n'
                    '- testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n'
                    'testfiles/select2/1/1sub1/1sub1sub2\n'
                    '     # Testing a full-line comment with leading and trailing spaces     \n'
                    'testfiles/select2/1/1sub2  \n'  # Note added spaces here
                    '+ testfiles/select2/1.py\n'
                    '+ testfiles/select2/3 \n'  # Note added space here
                    '+ testfiles/select2/1\n'
                    '- testfiles/select2/**')
        self.backup("full", "testfiles/select2", options=["--exclude-globbing-filelist=testfiles/exclude.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_exclude_filelist_trailing_whitespace_folders_work_with_quotes(self):
        """Test that folders with trailing whitespace in the names work correctly if they are enclosed in quotes"""
        # Create a filelist
        with open('testfiles/exclude.txt', 'w') as f:
            f.write('+ "testfiles/select2/trailing_space /trailing_space sub2/trailing_space sub2_file.txt"\n'  # New
                    '- "testfiles/select2/trailing_space /trailing_space sub2"\n'  # New
                    '+ "testfiles/select2/trailing_space "\n'  # New
                    '+ testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n'
                    'testfiles/select2/3/3sub3/3sub3sub2\n'
                    '+ testfiles/select2/3/3sub2/3sub2sub2\n'
                    '+ testfiles/select2/3/3sub3\n'
                    '- testfiles/select2/3/3sub1\n'
                    'testfiles/select2/2/2sub1/2sub1sub3\n'
                    'testfiles/select2/2/2sub1/2sub1sub2\n'
                    '+ testfiles/select2/2/2sub1\n'
                    'testfiles/select2/1/1sub3/1sub3sub2\n'
                    'testfiles/select2/1/1sub3/1sub3sub1\n'
                    'testfiles/select2/1/1sub2/1sub2sub3\n'
                    '+ testfiles/select2/1/1sub2/1sub2sub1\n'
                    'testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n'
                    'testfiles/select2/1/1sub1/1sub1sub2\n'
                    '- testfiles/select2/1/1sub2\n'
                    '+ testfiles/select2/1.py\n'
                    '+ testfiles/select2/3\n'
                    '+ testfiles/select2/1\n'
                    'testfiles/select2/**')
        self.backup("full", "testfiles/select2", options=["--exclude-filelist=testfiles/exclude.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree_with_trailing_space)

    def test_exclude_filelist_progress_option(self):
        """Test that exclude filelist is unaffected by the --progress option"""
        # Regression test for Bug #1264744 (https://bugs.launchpad.net/duplicity/+bug/1264744)
        # Create a filelist identical to that used in test_exclude_filelist
        with open('testfiles/exclude.txt', 'w') as f:
            f.write('+ testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n'
                    'testfiles/select2/3/3sub3/3sub3sub2\n'
                    '+ testfiles/select2/3/3sub2/3sub2sub2\n'
                    '+ testfiles/select2/3/3sub3\n'
                    '- testfiles/select2/3/3sub1\n'  # - added to ensure it makes no difference
                    'testfiles/select2/2/2sub1/2sub1sub3\n'
                    'testfiles/select2/2/2sub1/2sub1sub2\n'
                    '+ testfiles/select2/2/2sub1\n'
                    'testfiles/select2/1/1sub3/1sub3sub2\n'
                    'testfiles/select2/1/1sub3/1sub3sub1\n'
                    'testfiles/select2/1/1sub2/1sub2sub3\n'
                    '+ testfiles/select2/1/1sub2/1sub2sub1\n'
                    'testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n'
                    'testfiles/select2/1/1sub1/1sub1sub2\n'
                    '- testfiles/select2/1/1sub2\n'  # - added to ensure it makes no difference
                    '+ testfiles/select2/1.py\n'
                    '+ testfiles/select2/3\n'
                    '+ testfiles/select2/1\n'
                    'testfiles/select2/**')

        # Backup the files exactly as in test_exclude_filelist, but with the --progress option
        self.backup("full", "testfiles/select2", options=["--exclude-filelist=testfiles/exclude.txt",
                                                          "--progress"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        # The restored files should match those restored in test_exclude_filelist
        self.assertEqual(restored, self.expected_restored_tree)


class TestIncludeFilelistTest(IncludeExcludeFunctionalTest):
    """
    Test --include-filelist using duplicity binary.
    """

    def test_include_filelist(self):
        """Test that include filelist works in the basic case"""
        # See test_exclude_filelist above for explanation of what is expected. As this is an include filelist
        # any lines with no +/- modifier should be treated as if they have a +.
        # Create a filelist
        with open('testfiles/include.txt', 'w') as f:
            f.write('testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n'
                    '- testfiles/select2/3/3sub3/3sub3sub2\n'
                    'testfiles/select2/3/3sub2/3sub2sub2\n'
                    '+ testfiles/select2/3/3sub3\n'  # + added to ensure it makes no difference
                    '- testfiles/select2/3/3sub1\n'
                    '- testfiles/select2/2/2sub1/2sub1sub3\n'
                    '- testfiles/select2/2/2sub1/2sub1sub2\n'
                    'testfiles/select2/2/2sub1\n'
                    '- testfiles/select2/1/1sub3/1sub3sub2\n'
                    '- testfiles/select2/1/1sub3/1sub3sub1\n'
                    '- testfiles/select2/1/1sub2/1sub2sub3\n'
                    '+ testfiles/select2/1/1sub2/1sub2sub1\n'  # + added to ensure it makes no difference
                    '- testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n'
                    '- testfiles/select2/1/1sub1/1sub1sub2\n'
                    '- testfiles/select2/1/1sub2\n'
                    'testfiles/select2/1.py\n'
                    'testfiles/select2/3\n'
                    'testfiles/select2/1\n'
                    '- testfiles/select2/**')
        self.backup("full", "testfiles/select2", options=["--include-filelist=testfiles/include.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_include_filelist_combined_imperfections(self):
        """Test that include filelist works with imperfections in the input file"""
        # This is a combined test for speed reasons. The individual imperfections are tested as unittests in
        # unit/test_selection.
        # Imperfections tested are;
        # * Leading space/spaces before the modifier
        # * Trailing space/spaces after the filename (but before the newline)
        # * Blank lines (newline character only)
        # * Line only containing spaces
        # * Full-line comments with # as the first character and with leading/trailing spaces
        # * Unnecessarily quoted filenames with/without modifier  (both " and ')
        # Create a filelist
        with open('testfiles/include.txt', 'w') as f:
            f.write('testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n'
                    '- testfiles/select2/3/3sub3/3sub3sub2\n'
                    '"testfiles/select2/3/3sub2/3sub2sub2"\n'
                    '  + testfiles/select2/3/3sub3\n'  # + added to ensure it makes no difference
                    '- testfiles/select2/3/3sub1\n'
                    '- testfiles/select2/2/2sub1/2sub1sub3\n'
                    ' - "testfiles/select2/2/2sub1/2sub1sub2"\n'
                    'testfiles/select2/2/2sub1  \n'
                    '\n'
                    '- testfiles/select2/1/1sub3/1sub3sub2\n'
                    '- testfiles/select2/1/1sub3/1sub3sub1 \n'
                    "- 'testfiles/select2/1/1sub2/1sub2sub3'\n"
                    '             \n'
                    ' + testfiles/select2/1/1sub2/1sub2sub1 \n'  # + added to ensure it makes no difference
                    '- testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n'
                    '  - testfiles/select2/1/1sub1/1sub1sub2  \n'
                    '# Testing full-line comment\n'
                    '- testfiles/select2/1/1sub2\n'
                    "'testfiles/select2/1.py'\n"
                    'testfiles/select2/3\n'
                    '        #  Testing another full-line comment      \n'
                    'testfiles/select2/1\n'
                    '- testfiles/select2/**')
        self.backup("full", "testfiles/select2", options=["--include-filelist=testfiles/include.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_include_filelist_workaround_combined_imperfections_no_wildcards(self):
        """Test that include filelist works with imperfections in the input file"""
        # This is a modified version of test_include_filelist that passes, despite Bug #1408411
        # It is still a valid test, it just does not test as many selection features as the above.
        # This is a combined test for speed reasons. The individual imperfections are tested as unittests in
        # unit/test_selection.
        # Imperfections tested are;
        # * Leading space/spaces before the modifier
        # * Trailing space/spaces after the filename (but before the newline)
        # * Blank lines (newline character only)
        # * Line only containing spaces
        # * Full-line comments with # as the first character and with leading/trailing spaces
        # * Unnecessarily quoted filenames with/without modifier  (both " and ')
        # Create a filelist
        with open('testfiles/include.txt', 'w') as f:
            f.write('testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n'
                    'testfiles/select2/3/3sub2/3sub2sub2 \n'
                    '  + testfiles/select2/3/3sub3\n'  # + added to ensure it makes no difference
                    ' - testfiles/select2/3/3sub1  \n'
                    '- testfiles/select2/2/2sub1/2sub1sub3\n'
                    '- testfiles/select2/2/2sub1/2sub1sub2\n'
                    '"testfiles/select2/2/2sub1"\n'
                    '   - testfiles/select2/2/2sub3 \n'  # Added because of Bug #1408411
                    '- testfiles/select2/2/2sub2\n'  # Added because of Bug #1408411
                    "- 'testfiles/select2/1/1sub3/1sub3sub2'\n"
                    '\n'
                    '- testfiles/select2/1/1sub3/1sub3sub1\n'
                    '- testfiles/select2/1/1sub2/1sub2sub3\n'
                    '- "testfiles/select2/1/1sub2/1sub2sub2"\n'  # Added because of Bug #1408411
                    '# This is a full-line comment\n'
                    '+ testfiles/select2/1/1sub2/1sub2sub1  \n'  # + added to ensure it makes no difference
                    '- testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n'
                    '          \n'
                    '- testfiles/select2/1/1sub1/1sub1sub2\n'
                    #  '- testfiles/select2/1/1sub2\n'  # Commented out because of Bug #1408411
                    "'testfiles/select2/1.py'\n"
                    '       # This is another full-line comment, with spaces     \n'
                    'testfiles/select2/3\n'
                    #  '- testfiles/select2/2\n' # Commented out because of Bug #1408411
                    'testfiles/select2/1\n'
                    '- "testfiles/select2/trailing_space "\n'  # es instead of ea as no wildcard - **
                    '- testfiles/select2/1.doc')  # es instead of ea as no wildcard - **
        self.backup("full", "testfiles/select2", options=["--include-filelist=testfiles/include.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_include_globbing_filelist_combined_imperfections(self):
        """Test that include globbing filelist works with imperfections in the input file"""
        # Identical to test_include_filelist_combined_imperfections and included to ensure that
        # the deprecated --include-globbing-filelist function works as expected until it is deliberately removed.
        # This is a combined test for speed reasons. The individual imperfections are tested as unittests in
        # unit/test_selection.
        # Imperfections tested are;
        # * Leading space/spaces before the modifier
        # * Trailing space/spaces after the filename (but before the newline)
        # * Blank lines (newline character only)
        # * Line only containing spaces
        # * Full-line comments with # as the first character and with leading/trailing spaces
        # * Unnecessarily quoted filenames with/without modifier  (both " and ')
        # Create a filelist
        with open('testfiles/include.txt', 'w') as f:
            f.write('testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n'
                    '- testfiles/select2/3/3sub3/3sub3sub2\n'
                    '"testfiles/select2/3/3sub2/3sub2sub2"\n'
                    '  + testfiles/select2/3/3sub3\n'  # + added to ensure it makes no difference
                    '- testfiles/select2/3/3sub1\n'
                    '- testfiles/select2/2/2sub1/2sub1sub3\n'
                    ' - "testfiles/select2/2/2sub1/2sub1sub2"\n'
                    'testfiles/select2/2/2sub1  \n'
                    '\n'
                    '- testfiles/select2/1/1sub3/1sub3sub2\n'
                    '- testfiles/select2/1/1sub3/1sub3sub1 \n'
                    "- 'testfiles/select2/1/1sub2/1sub2sub3'\n"
                    '             \n'
                    ' + testfiles/select2/1/1sub2/1sub2sub1 \n'  # + added to ensure it makes no difference
                    '- testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n'
                    '  - testfiles/select2/1/1sub1/1sub1sub2  \n'
                    '# Testing full-line comment\n'
                    '- testfiles/select2/1/1sub2\n'
                    "'testfiles/select2/1.py'\n"
                    'testfiles/select2/3\n'
                    '        #  Testing another full-line comment      \n'
                    'testfiles/select2/1\n'
                    '- testfiles/select2/**')
        self.backup("full", "testfiles/select2", options=["--include-globbing-filelist=testfiles/include.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)


class TestIncludeExcludedForContents(IncludeExcludeFunctionalTest):
    """ Test to check that folders that are excluded are included if they contain includes of higher priority.
     Exhibits the issue reported in Bug #1408411 (https://bugs.launchpad.net/duplicity/+bug/1408411). """

    def write_filelist(self, filelist_name):
        """Used by the below tests to write the filelist"""
        assert filelist_name is not None
        with open(filelist_name, 'w') as f:
            f.write("+ testfiles/select/1/2/1\n"
                    "- testfiles/select/1/2\n"
                    "- testfiles/select/1/1\n"
                    "- testfiles/select/1/3")

    def restore_and_check(self):
        """Restores the backup and compares to what was expected (based on the filelist in write_filelist)"""
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [['2'], ['1']])

    def test_commandline_include_exclude(self):
        """test an excluded folder is included for included contents when using commandline includes and excludes"""
        self.backup("full", "testfiles/select/1",
                    options=["--include", "testfiles/select/1/2/1",
                             "--exclude", "testfiles/select/1/2",
                             "--exclude", "testfiles/select/1/1",
                             "--exclude", "testfiles/select/1/3"])
        self.restore_and_check()

    def test_include_globbing_filelist(self):
        """test an excluded folder is included for included contents with an include-globbing-filelist """
        # Deprecated, but include for now to ensure it keeps working until it is deliberately removed.
        self.write_filelist("testfiles/include.txt")
        self.backup("full", "testfiles/select/1", options=["--include-globbing-filelist=testfiles/include.txt"])
        self.restore_and_check()

    def test_exclude_globbing_filelist(self):
        """test an excluded folder is included for included contents with an exclude-globbing-filelist """
        # Deprecated, but include for now to ensure it keeps working until it is deliberately removed.
        self.write_filelist("testfiles/exclude.txt")
        self.backup("full", "testfiles/select/1", options=["--exclude-globbing-filelist=testfiles/exclude.txt"])
        self.restore_and_check()

    def test_include_filelist(self):
        """test an excluded folder is included for included contents with an include-filelist (non-globbing) """
        # Regression test for Bug #1408411 (https://bugs.launchpad.net/duplicity/+bug/1408411)
        self.write_filelist("testfiles/include.txt")
        self.backup("full", "testfiles/select/1", options=["--include-filelist=testfiles/include.txt"])
        self.restore_and_check()

    def test_exclude_filelist(self):
        """test an excluded folder is included for included contents with an exclude-filelist  (non-globbing) """
        # Regression test for Bug #1408411 (https://bugs.launchpad.net/duplicity/+bug/1408411)
        self.write_filelist("testfiles/exclude.txt")
        self.backup("full", "testfiles/select/1", options=["--exclude-filelist=testfiles/exclude.txt"])
        self.restore_and_check()


class TestAsterisks(IncludeExcludeFunctionalTest):
    """ Test to check that asterisks work as expected
     Exhibits the issue reported in Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371).
     See the unit tests for more granularity on the issue."""

    def restore_and_check(self):
        """Restores the backup and compares to what is expected."""
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [['2'], ['1']])

    def test_exclude_filelist_asterisks_none(self):
        """Basic exclude filelist."""
        with open("testfiles/filelist.txt", 'w') as f:
            f.write("+ testfiles/select/1/2/1\n"
                    "- testfiles/select/1/2\n"
                    "- testfiles/select/1/1\n"
                    "- testfiles/select/1/3")
        self.backup("full", "testfiles/select/1", options=["--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_exclude_filelist_asterisks_single(self):
        """Exclude filelist with asterisks replacing folders."""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        with open("testfiles/filelist.txt", 'w') as f:
            f.write("+ */select/1/2/1\n"
                    "- */select/1/2\n"
                    "- testfiles/*/1/1\n"
                    "- */*/1/3")
        self.backup("full", "testfiles/select/1", options=["--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_exclude_filelist_asterisks_double_asterisks(self):
        """Exclude filelist with double asterisks replacing folders."""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        with open("testfiles/filelist.txt", 'w') as f:
            f.write("+ **/1/2/1\n"
                    "- **/1/2\n"
                    "- **/select/1/1\n"
                    "- testfiles/select/1/3")
        self.backup("full", "testfiles/select/1", options=["--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_commandline_asterisks_single_excludes_only(self):
        """test_commandline_include_exclude with single asterisks on exclude lines."""
        self.backup("full", "testfiles/select/1",
                    options=["--include", "testfiles/select/1/2/1",
                             "--exclude", "testfiles/*/1/2",
                             "--exclude", "*/select/1/1",
                             "--exclude", "*/select/1/3"])
        self.restore_and_check()

    def test_commandline_asterisks_single_both(self):
        """test_commandline_include_exclude with single asterisks on both exclude and include lines."""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.backup("full", "testfiles/select/1",
                    options=["--include", "*/select/1/2/1",
                             "--exclude", "testfiles/*/1/2",
                             "--exclude", "*/select/1/1",
                             "--exclude", "*/select/1/3"])
        self.restore_and_check()

    def test_commandline_asterisks_double_exclude_only(self):
        """test_commandline_include_exclude with double asterisks on exclude lines."""
        self.backup("full", "testfiles/select/1",
                    options=["--include", "testfiles/select/1/2/1",
                             "--exclude", "**/1/2",
                             "--exclude", "**/1/1",
                             "--exclude", "**/1/3"])
        self.restore_and_check()

    def test_commandline_asterisks_double_both(self):
        """test_commandline_include_exclude with double asterisks on both exclude and include lines."""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.backup("full", "testfiles/select/1",
                    options=["--include", "**/1/2/1",
                             "--exclude", "**/1/2",
                             "--exclude", "**/1/1",
                             "--exclude", "**/1/3"])
        self.restore_and_check()

    def test_single_and_double_asterisks(self):
        """This compares a backup using --include-globbing-filelist with a single and double *."""
        with open("testfiles/filelist.txt", 'w') as f:
            f.write("+ testfiles/select2/*\n"
                    "- testfiles/select")
        self.backup("full", "testfiles/", options=["--include-globbing-filelist=testfiles/filelist.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir + "/select2")
        with open("testfiles/filelist2.txt", 'w') as f:
            f.write("+ testfiles/select2/**\n"
                    "- testfiles/select")
        self.backup("full", "testfiles/", options=["--include-globbing-filelist=testfiles/filelist2.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored2 = self.directory_tree_to_list_of_lists(restore_dir + "/select2")
        self.assertEqual(restored, restored2)

    def test_single_and_double_asterisks_includes_excludes(self):
        """This compares a backup using --includes/--excludes with a single and double *."""
        self.backup("full", "testfiles/",
                    options=["--include", "testfiles/select2/*",
                             "--exclude", "testfiles/select"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir + "/select2")
        self.backup("full", "testfiles/",
                    options=["--include", "testfiles/select2/**",
                             "--exclude", "testfiles/select"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored2 = self.directory_tree_to_list_of_lists(restore_dir + "/select2")
        self.assertEqual(restored, restored2)


class TestTrailingSlash(IncludeExcludeFunctionalTest):
    """ Test to check that a trailing slash works as expected
     Exhibits the issue reported in Bug #932482 (https://bugs.launchpad.net/duplicity/+bug/932482)."""

    def restore_and_check(self):
        """Restores the backup and compares to what is expected."""
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [['2'], ['1']])

    def test_exclude_filelist_trailing_slashes(self):
        """test_exclude_filelist_asterisks_none with trailing slashes."""
        with open("testfiles/filelist.txt", 'w') as f:
            f.write("+ testfiles/select/1/2/1/\n"
                    "- testfiles/select/1/2/\n"
                    "- testfiles/select/1/1/\n"
                    "- testfiles/select/1/3/")
        self.backup("full", "testfiles/select/1", options=["--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_exclude_filelist_trailing_slashes_single_wildcards_excludes(self):
        """test_exclude_filelist_trailing_slashes with single wildcards in excludes."""
        # Regression test for Bug #932482 (https://bugs.launchpad.net/duplicity/+bug/932482)
        with open("testfiles/filelist.txt", 'w') as f:
            f.write("+ testfiles/select/1/2/1/\n"
                    "- */select/1/2/\n"
                    "- testfiles/*/1/1/\n"
                    "- */*/1/3/")
        self.backup("full", "testfiles/select/1", options=["--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_exclude_filelist_trailing_slashes_double_wildcards_excludes(self):
        """test_exclude_filelist_trailing_slashes with double wildcards in excludes."""
        # Regression test for Bug #932482 (https://bugs.launchpad.net/duplicity/+bug/932482)
        with open("testfiles/filelist.txt", 'w') as f:
            f.write("+ testfiles/select/1/2/1/\n"
                    "- **/1/2/\n"
                    "- **/1/1/\n"
                    "- **/1/3/")
        self.backup("full", "testfiles/select/1", options=["--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_exclude_filelist_trailing_slashes_double_wildcards_excludes_2(self):
        """second test_exclude_filelist_trailing_slashes with double wildcards in excludes."""
        # Regression test for Bug #932482 (https://bugs.launchpad.net/duplicity/+bug/932482) and
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        with open("testfiles/filelist.txt", 'w') as f:
            f.write("+ **/1/2/1/\n"
                    "- **/1/2/\n"
                    "- **/1/1/\n"
                    "- **/1/3/")
        self.backup("full", "testfiles/select/1", options=["--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_exclude_filelist_trailing_slashes_wildcards(self):
        """test_commandline_asterisks_single_excludes_only with trailing slashes."""
        # Regression test for Bug #932482 (https://bugs.launchpad.net/duplicity/+bug/932482)
        self.backup("full", "testfiles/select/1",
                    options=["--include", "testfiles/select/1/2/1/",
                             "--exclude", "testfiles/*/1/2/",
                             "--exclude", "*/select/1/1/",
                             "--exclude", "*/select/1/3/"])
        self.restore_and_check()


class TestTrailingSlash2(IncludeExcludeFunctionalTest):
    """ This tests the behaviour of globbing strings with a trailing slash"""
    # See Bug #1479545 (https://bugs.launchpad.net/duplicity/+bug/1479545)

    def test_no_trailing_slash(self):
        """ Test that including 1.py works as expected"""
        self.backup("full", "testfiles/select2",
                    options=["--include", "testfiles/select2/1.py",
                             "--exclude", "**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [['1.py']])

    def test_trailing_slash(self):
        """ Test that globs with a trailing slash only match directories"""
        # Regression test for Bug #1479545
        # (https://bugs.launchpad.net/duplicity/+bug/1479545)
        self.backup("full", "testfiles/select2",
                    options=["--include", "testfiles/select2/1.py/",
                             "--exclude", "**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [])

    def test_include_files_not_subdirectories(self):
        """ Test that a trailing slash glob followed by a * glob only matches
        files and not subdirectories"""
        self.backup("full", "testfiles/select2",
                    options=["--exclude", "testfiles/select2/*/",
                             "--include", "testfiles/select2/*",
                             "--exclude", "**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [['1.doc', '1.py']])

    def test_include_subdirectories_not_files(self):
        """ Test that a trailing slash glob only matches directories"""
        self.backup("full", "testfiles/select2",
                    options=["--include", "testfiles/select2/1/1sub1/**/",
                             "--exclude", "testfiles/select2/1/1sub1/**",
                             "--exclude", "**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [['1'], ['1sub1'],
                                    ['1sub1sub1', '1sub1sub2', '1sub1sub3']])


class TestGlobbingReplacement(IncludeExcludeFunctionalTest):
    """ This tests the behaviour of the extended shell globbing pattern replacement functions."""
    # See the manual for a description of behaviours, but in summary:
    # * can be expanded to any string of characters not containing "/"
    # ? expands to any character except "/" and
    # [...] expands to a single character of those characters specified (ranges are acceptable).
    # The new special pattern, **, expands to any string of characters whether or not it contains "/".
    # Furthermore, if the pattern starts with "ignorecase:" (case insensitive), then this prefix will be
    # removed and any character in the string can be replaced with an upper- or lowercase version of itself.

    def test_globbing_replacement_in_includes(self):
        """ Test behaviour of the extended shell globbing pattern replacement functions in both include and exclude"""
        # Identical to test_include_exclude_basic with globbing characters added to both include and exclude lines
        # Exhibits the issue reported in Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371).
        # See above and the unit tests for more granularity on the issue.
        self.backup("full", "testfiles/select2",
                    options=["--include", "testfiles/select2/**/3sub3sub2/3sub3su?2_file.txt",  # Note ** and ? added
                             "--exclude", "testfiles/select2/*/3s*1",  # Note * added in both directory and filename
                             "--exclude", "testfiles/select2/**/2sub1sub3",  # Note ** added
                             "--exclude", "ignorecase:testfiles/select2/2/2sub1/2Sub1Sub2",  # Note ignorecase added
                             "--include", "ignorecase:testfiles/sel[w,u,e,q]ct2/2/2S?b1",    # Note ignorecase, [] and
                             # ? added
                             "--exclude", "testfiles/select2/1/1sub3/1s[w,u,p,q]b3sub2",  # Note [] added
                             "--exclude", "testfiles/select2/1/1sub[1-4]/1sub3sub1",  # Note [range] added
                             "--include", "testfiles/select2/*/1sub2/1s[w,u,p,q]b2sub1",  # Note * and [] added
                             "--exclude", "testfiles/select2/1/1sub1/1sub1sub3/1su?1sub3_file.txt",  # Note ? added
                             "--exclude", "testfiles/select2/1/1*1/1sub1sub2",  # Note * added
                             "--exclude", "testfiles/select2/1/1sub2",
                             "--include", "testfiles/select[2-4]/*.py",  # Note * and [range] added
                             "--include", "testfiles/*2/3",  # Note * added
                             "--include", "**/select2/1",  # Note ** added
                             "--exclude", "testfiles/select2/**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)


class TestExcludeIfPresent(IncludeExcludeFunctionalTest):
    """ This tests the behaviour of duplicity's --exclude-if-present option"""

    def test_exclude_if_present_baseline(self):
        """ Test that duplicity normally backs up files"""
        with open("testfiles/select2/1/1sub1/1sub1sub1/.nobackup", "w") as tag:
            tag.write("Files in this folder should not be backed up.")
        self.backup("full", "testfiles/select2/1/1sub1",
                    options=["--include", "testfiles/select2/1/1sub1/1sub1sub1/*",
                             "--exclude", "**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [['1sub1sub1'],
                                    ['.nobackup', '1sub1sub1_file.txt']])

    def test_exclude_if_present_excludes(self):
        """ Test that duplicity excludes files with relevant tag"""
        with open("testfiles/select2/1/1sub1/1sub1sub1/.nobackup", "w") as tag:
            tag.write("Files in this folder should not be backed up.")
        self.backup("full", "testfiles/select2/1/1sub1",
                    options=["--exclude-if-present", ".nobackup",
                             "--include", "testfiles/select2/1/1sub1/1sub1sub1/*",
                             "--exclude", "**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [])

    def test_exclude_if_present_excludes_2(self):
        """ Test that duplicity excludes files with relevant tag"""
        with open("testfiles/select2/1/1sub1/1sub1sub1/EXCLUDE.tag", "w") as tag:
            tag.write("Files in this folder should also not be backed up.")
        self.backup("full", "testfiles/select2/1/1sub1",
                    options=["--exclude-if-present", "EXCLUDE.tag",
                             "--include", "testfiles/select2/1/1sub1/1sub1sub1/*",
                             "--exclude", "**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [])


class TestLockedFoldersNoError(IncludeExcludeFunctionalTest):
    """ This tests that inaccessible folders do not cause an error"""

    @unittest.skipUnless(platform.platform().startswith('Linux'),
                         'Skip on non-Linux systems')
    def test_locked_baseline(self):
        """ Test no error if locked in path but excluded"""
        folder_to_lock = "testfiles/select2/1/1sub1/1sub1sub3"
        initial_mode = os.stat(folder_to_lock).st_mode
        os.chmod(folder_to_lock, 0o0000)
        self.backup("full", "testfiles/select2/1/1sub1",
                    options=["--include", "testfiles/select2/1/1sub1/1sub1sub1/*",
                             "--exclude", "**"])
        os.chmod(folder_to_lock, initial_mode)
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [['1sub1sub1'],
                                    ['1sub1sub1_file.txt']])

    @unittest.skipUnless(platform.platform().startswith('Linux'),
                         'Skip on non-Linux systems')
    def test_locked_excl_if_present(self):
        """ Test no error if excluded locked with --exclude-if-present"""
        # Regression test for Bug #1620085
        # https://bugs.launchpad.net/duplicity/+bug/1620085
        folder_to_lock = "testfiles/select2/1/1sub1/1sub1sub3"
        initial_mode = os.stat(folder_to_lock).st_mode
        os.chmod(folder_to_lock, 0o0000)
        self.backup("full", "testfiles/select2/1/1sub1",
                    options=["--exclude-if-present", "EXCLUDE.tag",
                             "--include", "testfiles/select2/1/1sub1/1sub1sub1/*",
                             "--exclude", "**"])
        os.chmod(folder_to_lock, initial_mode)
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [['1sub1sub1'],
                                    ['1sub1sub1_file.txt']])


class TestFolderIncludesFiles(IncludeExcludeFunctionalTest):
    """ This tests that including a folder includes the files within it"""
    # https://bugs.launchpad.net/duplicity/+bug/1624725

    def test_includes_files(self):
        """This tests that including a folder includes the files within it"""
        self.backup("full", "testfiles/select2/1/1sub1",
                    options=["--include", "testfiles/select2/1/1sub1/1sub1sub1",
                             "--exclude", "**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [['1sub1sub1'],
                                    ['1sub1sub1_file.txt']])

    def test_includes_files_trailing_slash(self):
        """This tests that including a folder includes the files within it"""
        self.backup("full", "testfiles/select2/1/1sub1",
                    options=["--include", "testfiles/select2/1/1sub1/1sub1sub1/",
                             "--exclude", "**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [['1sub1sub1'],
                                    ['1sub1sub1_file.txt']])

    def test_includes_files_trailing_slash_globbing_chars(self):
        """Tests folder includes with globbing char and /"""
        self.backup("full", "testfiles/select2/1/1sub1",
                    options=["--include", "testfiles/s?lect2/1/1sub1/1sub1sub1/",
                             "--exclude", "**"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [['1sub1sub1'],
                                    ['1sub1sub1_file.txt']])

    def test_excludes_files_no_trailing_slash(self):
        """This tests that excluding a folder excludes the files within it"""
        self.backup("full", "testfiles/select2/1/1sub1",
                    options=["--exclude", "testfiles/select2/1/1sub1/1sub1sub1",
                             "--exclude", "testfiles/select2/1/1sub1/1sub1sub2",
                             "--exclude", "testfiles/select2/1/1sub1/1sub1sub3",
                             "--include", "testfiles/select2/1/1sub1/1sub1**",
                             "--exclude", "testfiles/select2/1/1sub1/irrelevant.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [])

    def test_excludes_files_trailing_slash(self):
        """Excluding a folder excludes the files within it, if ends with /"""
        self.backup("full", "testfiles/select2/1/1sub1",
                    options=["--exclude", "testfiles/select2/1/1sub1/1sub1sub1/",
                             "--exclude", "testfiles/select2/1/1sub1/1sub1sub2/",
                             "--exclude", "testfiles/select2/1/1sub1/1sub1sub3/",
                             "--include", "testfiles/select2/1/1sub1/1sub1**",
                             "--exclude", "testfiles/select2/1/1sub1/irrelevant.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [])

    def test_excludes_files_trailing_slash_globbing_chars(self):
        """Tests folder excludes with globbing char and /"""
        self.backup("full", "testfiles/select2/1/1sub1",
                    options=["--exclude", "testfiles/sel?ct2/1/1sub1/1sub1sub1/",
                             "--exclude", "testfiles/sel[e,f]ct2/1/1sub1/1sub1sub2/",
                             "--exclude", "testfiles/sel*t2/1/1sub1/1sub1sub3/",
                             "--include", "testfiles/select2/1/1sub1/1sub1**",
                             "--exclude", "testfiles/select2/1/1sub1/irrelevant.txt"])
        self.restore()
        restore_dir = 'testfiles/restore_out'
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [])

if __name__ == "__main__":
    unittest.main()
